#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import http
import json
import unittest

import django.test
import django.conf
import django.core.exceptions
import mock

from django_declarative_apis import models
from django_declarative_apis.resources import resource
from django_declarative_apis.authentication.oauthlib import oauth_errors
from tests import testutils


class ResourceTestCase(testutils.RequestCreatorMixin, django.test.TestCase):
    def test_call_post(self):
        def handle_post(request, *args, **kwargs):
            return http.HTTPStatus.OK, {'foo': 'bar'}

        class Handler:
            allowed_methods = ('POST',)
            method_handlers = {
                'POST': handle_post
            }

        body = {'foo': 'bar'}
        req = self.create_request(method='POST', body=body)

        res = resource.Resource(lambda: Handler())
        resp = res(req)
        self.assertEqual(json.loads(resp.content), body)

    def test_call_invalid_mime_type(self):
        class Handler:
            allowed_methods = ('POST',)
            method_handlers = {
                'POST': lambda req, *args, **kwargs: (http.HTTPStatus.OK, '')
            }

        body = {'foo': 'bar'}
        req = self.create_request(method='POST', body=body)

        res = resource.Resource(lambda: Handler())
        with mock.patch('django_declarative_apis.resources.resource.translate_mime') as mock_translate:
            mock_translate.side_effect = resource.MimerDataException

            resp = res(req)

    def test_call_put(self):
        class Handler:
            allowed_methods = ('PUT',)
            method_handlers = {
                'PUT': lambda req, *args, **kwargs: (http.HTTPStatus.OK, '')
            }

        body = {'foo': 'bar'}
        req = self.create_request(method='PUT', body=body)
        res = resource.Resource(lambda: Handler())
        resp = res(req)

        # make sure request coercion did its thing to allow django to support it
        self.assertEqual(req.PUT, req.POST)

    def test_anonymous(self):
        class HandlerA:
            anonymous = lambda: True

        handler = HandlerA()
        res = resource.Resource(lambda: handler)
        self.assertEqual(res.anonymous, handler.anonymous)
        
        class HandlerB:
            pass

        handler = HandlerB()
        res = resource.Resource(lambda: handler)
        self.assertIsNone(res.anonymous)

    def test_authenticate(self):
        class Handler:
            pass

        handler = Handler()
        res = resource.Resource(lambda: handler)
        _, anonymous, err = res.authenticate(django.test.RequestFactory().get('/'), 'GET')
        self.assertEqual(anonymous, resource.CHALLENGE)
        self.assertIsInstance(err, oauth_errors.OAuthMissingParameterError)

    def test_deserializers(self):
        data = {'foo': 'bar'}
        req = self.create_request(method='POST', body=data, content_type='application/json')
        resource._deserialize_json(req)
        self.assertEqual(req.POST, data)

    def test_init_unset_authentication_handlers(self):
        current_handlers = django.conf.settings.DECLARATIVE_ENDPOINT_AUTHENTICATION_HANDLERS
        try:
            del resource.settings.DECLARATIVE_ENDPOINT_AUTHENTICATION_HANDLERS
            self.assertRaises(
                django.core.exceptions.ImproperlyConfigured,
                resource.Resource, lambda: lambda: None
            )
        finally:
            resource.settings.DECLARATIVE_ENDPOINT_AUTHENTICATION_HANDLERS = current_handlers

    def test_init_uncallable_handler(self):
        try:
            resource.Resource('not callable')
            self.fail('should have faileid')
        except AttributeError as err:
            self.assertEqual(str(err), 'Handler not callable.')

    def test_error_handler(self):
        class Handler:
            pass

        res = resource.Resource(lambda: Handler())
        req = django.test.RequestFactory().get('/')
        try:
            raise Exception('something bad happened')
        except Exception as err:
            with mock.patch('django_declarative_apis.resources.resource.Resource.email_exception') as mock_email:
                resp = res.error_handler(err, req, 'GET', 'json')

        (reporter,), kwargs = mock_email.call_args_list[0]
        self.assertEqual(mock_email.call_count, 1)
        self.assertEqual(reporter.request, req)
        self.assertEqual(reporter.exc_type, Exception)

    def test_use_emitter(self):
        resp = django.http.HttpResponse(content_type='image')
        self.assertFalse(resource.Resource._use_emitter(resp))

        resp = django.http.HttpResponse(content_type='application/json')
        self.assertFalse(resource.Resource._use_emitter(resp))

        resp = django.http.HttpResponse(content_type='unhandled')
        self.assertTrue(resource.Resource._use_emitter(resp))

    def test_email_exception(self):
        with mock.patch('django_declarative_apis.resources.resource.EmailMessage') as mock_email:
            traceback = 'something bad happened'
            reporter = type('Reporter', (), {'get_traceback_html': lambda: traceback})

            res = resource.Resource(lambda: lambda: None)
            res.email_exception(reporter)
            (subj, body, _, __), ___ = mock_email.call_args_list[0]
            self.assertEqual(subj, '[Django] Django Declarative APIs crash report')
            self.assertEqual(body, traceback)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import http
import json
import unittest

import mock
import django.test

from django_declarative_apis.resources import utils

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization

from tests import testutils


class UtilsTestCase(unittest.TestCase):

    def test_format_error(self):
        val = utils.format_error(Exception('something bad happened'))
        self.assertIn('something bad happened', val)

    def test_rc_factory_invalid_code(self):
        try:
            utils.rc.INVALID
            self.fail('should have failed')
        except AttributeError:
            pass

    def test_rc_factory_iterable_content(self):
        data = [b'foo', b'bar']
        resp = utils.rc.NOT_FOUND
        resp.content = data
        self.assertEqual(resp.content, b''.join(data))

    def test_form_validation_error(self):
        err = utils.FormValidationError('not a form')
        self.assertEqual(err.form, 'not a form')

    def test_require_mime(self):
        @utils.require_mime('application/json')
        def test_json(self, req):
            return {}

        self.assertEqual(
            test_json(
                self,
                django.test.RequestFactory().post('/', {'foo': 'bar'}, content_type='application/json')
            ), {}
        )

        self.assertEqual(
            test_json(
                self,
                django.test.RequestFactory().post('/', '', content_type='text/plain')
            ).status_code, http.HTTPStatus.BAD_REQUEST
        )


class MimerTestCase(unittest.TestCase):
    def test_is_multipart(self):
        req = django.test.RequestFactory().get('/', content_type='text/plain')
        self.assertFalse(utils.Mimer(req).is_multipart())

        req = django.test.RequestFactory().post('/', content_type='multipart/form-data')
        self.assertTrue(utils.Mimer(req).is_multipart())

    def test_loader_for_type(self):
        req = django.test.RequestFactory().post('/', json.dumps({'foo': 'bar'}), content_type='application/json')
        mimer = utils.Mimer(req)
        self.assertEqual(mimer.loader_for_type('application/json'), json.loads)

    def test_translate(self):
        data = {'foo': 'bar'}
        req = django.test.RequestFactory().post('/', json.dumps(data), content_type='application/json')
        mimer = utils.Mimer(req)
        req = mimer.translate()
        self.assertEqual(req.POST, {})
        self.assertEqual(req.PUT, {})
        self.assertEqual(req.data, data)

    def test_translate_none_deserializer(self):
        with mock.patch.dict('django_declarative_apis.resources.utils.Mimer.TYPES', {None: 'application_json'}, clear=True):
            req = django.test.RequestFactory().post('/', {'foo': 'bar'}, content_type='application/json')
            mimer = utils.Mimer(req)
            self.assertIsNone(mimer.translate().data)

    def test_translate_invalid_data(self):
        req = django.test.RequestFactory().post('/', 'notjson', content_type='application/json')
        mimer = utils.Mimer(req)
        self.assertRaises(utils.MimerDataException, mimer.translate)

    def test_register_unregister(self):
        deserializer, content_type = lambda _: None, 'foo'
        utils.Mimer.register(deserializer, content_type)

        req = django.test.RequestFactory().post('/', '', content_type=content_type)
        mimer = utils.Mimer(req)
        self.assertEqual(mimer.loader_for_type(content_type), deserializer)

        utils.Mimer.unregister(deserializer)
        self.assertIsNone(mimer.loader_for_type(content_type))


class KeyProcessingTestCase(unittest.TestCase):

    # We can't compute these keys from the cryptography library because they're
    # nonstandard. These were both taken from their respective devices.

    KEYS = {
        'ANDROID_PUBLIC': (
            '-----BEGIN RSA PUBLIC KEY-----\n'
            'MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQC0YjCwIfYoprq/FQO6lb3asXrx\n'
            'LlJFuCvtinTF5p0GxvQGu5O3gYytUvtC2JlYzypSRjVxwxrsuRcP3e641SdASwfr\n'
            'mzyvIgP08N4S0IFzEURkV1wp/IpH7kH41EtbmUmrXSwfNZsnQRE5SYSOhh+LcK2w\n'
            'yQkdgcMv11l4KoBkcwIDAQAB\n'
            '-----END RSA PUBLIC KEY-----'
            ),
        'IOS_PUBLIC': (
            '-----BEGIN CERTIFICATE-----\n'
            'MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEArZvjX99s6wyS3IKdmPYOEZ'
            'm0CeFIyWwl/S1D7PyOfJkd3bHFzMrRLbgB7gXubZWxiORQJKAW4gfDpBXq0kes8LEhV'
            'gVdMdmvuCyGqchC5ON8hMXrrycQa8ejr0eq81y+bc1cYCE8mtVQqXdgrTYQAN0ObIkJ'
            'ajzfiQ9KOCoszopZoNoDP9qPYTtlURQvR5VwWeAGmL9ZkqdQqBr2tMlV673kdXo1HuB'
            '97K9mcYLBN8a5mQ/YgdzjPf0+JjpSH/Fx9ElCVJP8zZbBW2/06SMs3xDLCTHmdRcs/o'
            'WLhoyYRMvagq/j0MBNP2fU9dtlqPLFl6ZjJl/M4JdeZJWURkI9ZQIDAQAB\n'
            '-----END CERTIFICATE-----'
            ),
    }

    def test_nonstandard_public_key(self):
        for key_name in ['IOS_PUBLIC', 'ANDROID_PUBLIC']:
            with self.subTest(key_name=key_name):
                processed = utils.preprocess_rsa_key(self.KEYS[key_name])
                self.assertTrue(processed.startswith('-----BEGIN PUBLIC KEY-----'))
                self.assertTrue(processed.endswith('-----END PUBLIC KEY-----'))

    def test_standard_public_key(self):
        # example from `cryptography` documentation
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend())
        public_key = private_key.public_key()
        public_key_str = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo).decode('utf8')
        self.assertEqual(public_key_str, utils.preprocess_rsa_key(public_key_str))



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import http
import json
import unittest

import django.test
import django.http
import mock

from django_declarative_apis import models
from django_declarative_apis import authentication
from django_declarative_apis.authentication.oauthlib import oauth1, oauth_errors
from tests import testutils


class TwoLeggedOauth1TestCase(unittest.TestCase):
    def setUp(self):
        self.request_factory = django.test.RequestFactory()
        self.consumer = models.OauthConsumer.objects.create(name='test')

    def test_validate_missing_parameters(self):
        request = self.request_factory.get('/')
        authenticator = oauth1.TwoLeggedOauth1()
        authenticator.validate_missing_parameters(request)

        self.assertEqual(
            request.auth_header,
            'OAuth realm="API",oauth_problem=parameter_absent&oauth_parameters_absent=oauth_consumer_key,oauth_nonce,'
            'oauth_signature,oauth_signature_method,oauth_timestamp'
        )

    def test_is_authenticated_django_header_sillyness_and_auth_failure(self):
        request = self.request_factory.get('/')
        request.META['HTTP_AUTHORIZATION'] = 'OAuth foo=bar'
        testutils.OAuthClientHandler._build_request(request)

        authenticator = oauth1.TwoLeggedOauth1()
        result = authenticator.is_authenticated(request)
        self.assertIsInstance(result, oauth_errors.OAuthMissingParameterError)
        self.assertEqual(
            result.detail,
            'Parameters missing: oauth_consumer_key,oauth_nonce,oauth_signature,oauth_signature_method,oauth_timestamp'
        )

    def test_is_authenticated_invalid_signature_error(self):
        request = self.request_factory.get('/')
        request.META['consumer'] = self.consumer
        testutils.OAuthClientHandler._build_request(request)
        
        params = request.GET.copy()
        params['oauth_signature'] = 'bad'
        request.GET = params
        
        authenticator = oauth1.TwoLeggedOauth1()
        result = authenticator.is_authenticated(request)
        self.assertIsInstance(result, oauth_errors.OAuthInvalidSignatureError)
        self.assertTrue(result.detail.startswith('Invalid signature.'))

    @mock.patch('django_declarative_apis.authentication.oauthlib.oauth1.logging.error')
    def test_is_authenticated_validation_error_handled(self, mocked_log):
        class ExceptionWithMessage(Exception):
            def __init__(self, msg):
                self.message = msg


        request = self.request_factory.get('/')
        request.META['consumer'] = self.consumer
        testutils.OAuthClientHandler._build_request(request)

        authenticator = oauth1.TwoLeggedOauth1()

        with mock.patch('django_declarative_apis.authentication.oauthlib.oauth1.TweakedSignatureOnlyEndpoint') as mocked_obj:
            mocked_obj.side_effect = ExceptionWithMessage(msg='something bad happened')
            result = authenticator.is_authenticated(request)
        self.assertIsInstance(result, authentication.AuthenticationFailure)
        mocked_log.assert_called_with('Invalid oauthlib request: something bad happened')

    def test_authenticate_header(self):
        request = self.request_factory.get('/')

        authenticator = oauth1.TwoLeggedOauth1()

        result = authenticator.authenticate_header(request)
        self.assertEqual(result, 'Unknown OAuth Error')

        request.auth_header = 'foo'
        result = authenticator.authenticate_header(request)
        self.assertEqual(result, 'foo')

    def test_challenge_with_no_error(self):
        authenticator = oauth1.TwoLeggedOauth1()
        response = authenticator.challenge()

        self.assertIsInstance(response, django.http.HttpResponse) 
        self.assertEqual(response.status_code, http.HTTPStatus.UNAUTHORIZED)
        self.assertEqual(response['www-authenticate'], 'None')

    def test_challenge_with_error(self):
        authenticator = oauth1.TwoLeggedOauth1()
        response = authenticator.challenge(oauth_errors.OAuthTimestampError())

        self.assertEqual(response.status_code, http.HTTPStatus.UNAUTHORIZED)
        self.assertTrue('timestamp_refused' in response['www-authenticate'])

        data = json.loads(response.getvalue().decode('utf8'))
        self.assertEqual(data['error_code'], http.HTTPStatus.UNAUTHORIZED)
        self.assertTrue(data['error_message'], 'There was a problem with your timestamp.')



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import logging
from pydoc import locate

from django.core.cache import cache


def rate_limit_exceeded(key='', timeout=1):
    """
    :param key: rate limit bin
    :param timeout: number of seconds to enforce between requests
    :return: True if rate limit has been exceeded (request should be throttled)

    The maximum supported rate limit is 1 request/second

    Bins should be narrow enough that multiple requests per second are not
    necessary (i.e. a requester could make many requests per second, but we
    would expect that requests related to any given RequesterUser are limited
    to a max of 1 per 5 seconds or something).

    The options for supporting rate limits higher than 1/second would all
    involve making multiple redis calls per request to enforce the rate limit,
    and would require more complex data structures in redis.

    This is not intended to be used for denial-of-service protection - that is
    frankly too complex for our API, and should probably come from a proxy.
    The design intent of this feature is protecting OTP authentication against
    brute-force attacks.
    """
    result = not cache.add('ratelimit:{0}'.format(key), '1', timeout=timeout)
    if result:
        logging.warning("Detected rate limit exceeded! key='%s'", key)

    return result

def locate_object(namespaced_name):
    name_components = namespaced_name.split('.')
    module_name = '.'.join(name_components[:-1])
    module = locate(module_name)
    class_name = name_components[-1]
    return getattr(module, class_name)


def instantiate_class(namespaced_class_name, *args, **kwargs):
    cls = locate_object(namespaced_class_name)
    return cls(*args, **kwargs)


#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import decimal, re, inspect
import copy

from django.conf import settings

try:
    # yaml isn't standard with python.  It shouldn't be required if it
    # isn't used.
    import yaml
except ImportError:  # pragma: nocover
    yaml = None

from django.db import models
from django.utils.xmlutils import SimplerXMLGenerator
from django.utils.encoding import smart_str
from django.urls import reverse, NoReverseMatch
try:
    from django.core.serializers.json import DateTimeAwareJSONEncoder as DjangoJSONEncoder
except ImportError:
    from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse
from django.core import serializers

import json
import types

from .utils import HttpStatusCode, Mimer, locate_object

from io import StringIO

import pickle

# Allow people to change the reverser (default `permalink`).
reverser = models.permalink




class Emitter(object):
    """
    Super emitter. All other emitters should subclass
    this one. It has the `construct` method which
    conveniently returns a serialized `dict`. This is
    usually the only method you want to use in your
    emitter. See below for examples.

    `RESERVED_FIELDS` was introduced when better resource
    method detection came, and we accidentially caught these
    as the methods on the handler. Issue58 says that's no good.
    """
    EMITTERS = {}
    RESERVED_FIELDS = {'read', 'update', 'create', 'delete', 'model', 'anonymous', 'allowed_methods', 'fields',
                       'exclude'}

    def __init__(self, payload, handler, anonymous=True):
        self.data = payload
        self.handler = handler
        self.anonymous = anonymous

        if isinstance(self.data, Exception):
            raise Exception from self.data

    def method_fields(self, handler, fields):
        if not handler:
            return {}

        ret = dict()

        for field in fields - Emitter.RESERVED_FIELDS:
            t = getattr(handler, str(field), None)

            if t and callable(t):
                ret[field] = t

        return ret

    def construct(self):
        if isinstance(self.data, HttpResponse):
            raise HttpStatusCode(self.data)
        else:
            return self.data

    def render(self, request):
        """
        This super emitter does not implement `render`,
        this is a job for the specific emitter below.
        """
        raise NotImplementedError("Please implement render.")

    def stream_render(self, request, stream=True):
        """
        Tells our patched middleware not to look
        at the contents, and returns a generator
        rather than the buffered string. Should be
        more memory friendly for large datasets.
        """
        yield self.render(request)

    @classmethod
    def get(cls, format):
        """
        Gets an emitter, returns the class and a content-type.
        """
        if format in cls.EMITTERS:
            return cls.EMITTERS.get(format)

        raise ValueError("No emitters found for type %s" % format)

    @classmethod
    def register(cls, name, klass, content_type='text/plain'):
        """
        Register an emitter.

        Parameters::
         - `name`: The name of the emitter ('json', 'xml', 'yaml', ...)
         - `klass`: The emitter class.
         - `content_type`: The content type to serve response as.
        """
        cls.EMITTERS[name] = (klass, content_type)

    @classmethod
    def unregister(cls, name):
        """
        Remove an emitter from the registry. Useful if you don't
        want to provide output in one of the built-in emitters.
        """
        return cls.EMITTERS.pop(name, None)


class XMLEmitter(Emitter):
    def _to_xml(self, xml, data):
        if isinstance(data, (list, tuple)):
            for item in data:
                xml.startElement("resource", {})
                self._to_xml(xml, item)
                xml.endElement("resource")
        elif isinstance(data, dict):
            for key, value in data.items():
                xml.startElement(key, {})
                self._to_xml(xml, value)
                xml.endElement(key)
        else:
            xml.characters(smart_str(data))

    def render(self, request):
        stream = StringIO()

        xml = SimplerXMLGenerator(stream, "utf-8")
        xml.startDocument()
        xml.startElement("response", {})

        self._to_xml(xml, self.construct())

        xml.endElement("response")
        xml.endDocument()

        return stream.getvalue()


Emitter.register('xml', XMLEmitter, 'text/xml; charset=utf-8')
Mimer.register(lambda *a: None, ('text/xml',))


class JSONEmitter(Emitter):
    """
    JSON emitter, understands timestamps.
    """

    def decode(self, data):
        if isinstance(data, list):
            for num, val in enumerate(data):
                try:
                    data[num] = val.decode('utf8')
                except AttributeError:
                    pass
        return data

    def render(self, request):
        cb = request.GET.get('callback', None)
        assert cb is None, 'JSONP Callbacks not suppoted'
        seria = json.dumps(
            self.decode(self.construct()), cls=DjangoJSONEncoder, ensure_ascii=False, indent=4
        )

        # Callback
        # TODO: do we care about JSONP?
        # if cb and is_valid_jsonp_callback_value(cb):
        #     return '%s(%s)' % (cb, seria)

        return seria


Emitter.register('json', JSONEmitter, 'application/json; charset=utf-8')
Mimer.register(json.loads, ('application/json',))


class YAMLEmitter(Emitter):
    """
    YAML emitter, uses `safe_dump` to omit the
    specific types when outputting to non-Python.
    """

    def render(self, request):
        return yaml.safe_dump(self.construct())


if yaml:  # Only register yaml if it was import successfully.
    Emitter.register('yaml', YAMLEmitter, 'application/x-yaml; charset=utf-8')
    Mimer.register(lambda s: dict(yaml.safe_load(s)), ('application/x-yaml',))


class PickleEmitter(Emitter):
    """
    Emitter that returns Python pickled.
    """

    def render(self, request):
        return pickle.dumps(self.construct())


Emitter.register('pickle', PickleEmitter, 'application/python-pickle')

"""
WARNING: Accepting arbitrary pickled data is a huge security concern.
The unpickler has been disabled by default now, and if you want to use
it, please be aware of what implications it will have.

Read more: http://nadiana.com/python-pickle-insecure

Uncomment the line below to enable it. You're doing so at your own risk.
"""


# Mimer.register(pickle.loads, ('application/python-pickle',))

class DjangoEmitter(Emitter):
    """
    Emitter for the Django serialized format.
    """

    def render(self, request, format='xml'):
        if isinstance(self.data, HttpResponse):
            return self.data
        elif isinstance(self.data, (int, str)):
            response = self.data
        else:
            response = serializers.serialize(format, self.data, indent=True)

        return response


Emitter.register('django', DjangoEmitter, 'text/xml; charset=utf-8')



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import json
import unittest
import pickle
from xml.dom import minidom

import django.test
import django.http
import yaml
from django.db import models

from django_declarative_apis.resources import emitters
import tests.models


class EmitterTestCase(unittest.TestCase):
    def test_init_data_exception_error(self):
        class TestException(Exception):
            pass

        try:
            raise TestException('something bad happened')
        except TestException as err:
            try:
                emitters.Emitter(err, None)
            except Exception as reraised_err:
                # make sure the new exception is being re-raised from the initial one
                self.assertEqual(reraised_err.__traceback__.tb_frame, err.__traceback__.tb_frame)

    def test_method_fields(self):
        class _Handler:
            def test(self):
                pass

            def __call__(self):
                pass

        handler = _Handler()
        em = emitters.Emitter(None, handler)
        fields = em.method_fields(handler, {'test',})
        self.assertEqual(fields['test'], handler.test)

        em = emitters.Emitter(None, None)
        self.assertEqual(em.method_fields(None, {'test'}), {})

    def test_render_raises_not_implemented_error(self):
        em = emitters.Emitter(None, None)
        self.assertRaises(NotImplementedError, em.render, None)

    def test_stream_render(self):
        data = ['foo', 'bar']
        class _StreamEmitter(emitters.Emitter):
            def __init__(self, _, __):
                self.idx = 0

            def render(self, request):
                try:
                    val = data[self.idx]
                except IndexError:
                    raise StopIteration

                self.idx += 1
                return val

        em = _StreamEmitter(None, None)
        for idx, val in enumerate(em.stream_render(django.test.RequestFactory().get('/'))):
            self.assertEqual(val, data[idx])

    def test_register_unregister_emitters(self):
        class FooEmitter:
            pass

        emitters.Emitter.register('foo', FooEmitter())
        emitter, content_type = emitters.Emitter.get('foo')
        self.assertIsInstance(emitter, FooEmitter)
        self.assertEqual(content_type, 'text/plain')
        emitters.Emitter.unregister('foo')
        self.assertRaises(ValueError, emitters.Emitter.get, 'foo')


class XMLEmitterTestCase(unittest.TestCase):
    def test_render_success(self):
        em = emitters.XMLEmitter({'foo': 'bar'}, lambda: None)
        resp = em.render(django.test.RequestFactory().get('/'))
        dom = minidom.parseString(resp)
        self.assertEqual(
            dom.getElementsByTagName('foo')[0].childNodes[0].data,
            'bar'
        )

    def test_render_list(self):
        em = emitters.XMLEmitter(['foo', 'bar'], lambda: None)
        resp = em.render(django.test.RequestFactory().get('/'))
        dom = minidom.parseString(resp)
        self.assertEqual(
            len(dom.getElementsByTagName('resource')),
            2
        )


class JSONEmitterTestCase(unittest.TestCase):
    def test_decode(self):
        # one bytes object and one string to test internal decoding
        em = emitters.JSONEmitter([b'foo', 'bar'], lambda: None)
        resp = em.render(django.test.RequestFactory().get('/'))
        self.assertEqual(json.loads(resp), ['foo', 'bar'])


class DjangoEmitterTestCase(unittest.TestCase):
    def test_render_http_response_succes(self):
        class _Handler:
            def __call__(self):
                pass

        resp = django.http.HttpResponse(content=b'foo')
        em = emitters.DjangoEmitter(resp, _Handler())
        self.assertEqual(em.render(None), resp)

    def test_render_str_int_success(self):
        data = 'string data'
        em = emitters.DjangoEmitter(data, lambda: None)
        self.assertEqual(em.render(None), data)

        data = 42
        em = emitters.DjangoEmitter(data, lambda: None)
        self.assertEqual(em.render(None), data)

    def test_render_serializable_success(self):
        tests.models.TestModel.objects.all().delete()
        tests.models.TestModel()

        em = emitters.DjangoEmitter(tests.models.TestModel.objects.all(), lambda: None)
        resp = em.render(None, format='json')
        self.assertEqual(json.loads(resp), [])

        obj = tests.models.TestModel(int_field=42)
        obj.save()

        em = emitters.DjangoEmitter(tests.models.TestModel.objects.all(), lambda: None)
        resp = em.render(None, format='json')
        self.assertEqual(
            json.loads(resp),
            [{
                'model': 'tests.testmodel',
                'pk': obj.pk,
                'fields': {
                    'int_field': 42,
                }
            }]
        )


class YAMLEmitterTestCase(unittest.TestCase):
    def test_render(self):
        data = {'foo': 'bar'}
        em = emitters.YAMLEmitter(data, lambda: None)
        self.assertEqual(yaml.load(em.render(None)), data)


class PickleEmitterTestCase(unittest.TestCase):
    def test_render(self):
        data = {'foo': 'bar'}
        em = emitters.PickleEmitter(data, lambda: None)
        self.assertEqual(pickle.loads(em.render(None)), data)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

# pragma: nocover
import pprint

from django.conf import settings
from django.core.management.base import BaseCommand
from django.core.urlresolvers import RegexURLResolver, RegexURLPattern

from django_declarative_apis.resources.resource import Resource


class Command(BaseCommand):
    help = 'Generate documentation for declarative endpoints'

    def handle(self, *args, **options):
        pprint.pprint(self.traverse())

    def traverse(self, url_patterns=None, prefix=None):
        if url_patterns is None:
            url_patterns = __import__(settings.ROOT_URLCONF).urls.urlpatterns
        if prefix is None:
            prefix = ''

        result = []
        for url_pattern in url_patterns:
            endpoint_location = prefix + url_pattern.regex.pattern
            if isinstance(url_pattern, RegexURLResolver):
                result += self.traverse(url_pattern.url_patterns, endpoint_location)
            elif isinstance(url_pattern, RegexURLPattern):
                callback = url_pattern.callback
                if isinstance(callback, Resource):
                    handler = callback.handler
                    result.append({'location': endpoint_location, 'endpoint': handler.documentation})
        return result



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import http
import uuid
import logging
import warnings

import http.client

# Start Toopher-specific codes at 600 to avoid conflict/confusion with HTTP status codes
import sys

HTTPS_REQUIRED = (600, "This request must use HTTPS")
FORBIDDEN = (601, "Not allowed")
EXTERNAL_REQUEST_FAILURE = (604, "Encountered exception while contacting an external service")
REQUEST_THROTTLED = (605, "Too many requests.  Please try again later.")
TIMED_OUT = (606, "Request timed out")
AUTHORIZATION_FAILURE = (607, "Authorization failure")

EXTRA_FIELDS = (700, "Extra field(s) in request")
READ_ONLY_FIELDS = (701, "Write attempted to read-only field(s)")
MISSING_FIELDS = (702, "Missing required field(s)")
INVALID_FIELD_VALUES = (703, "Invalid values for field(s)")

LOGGED_SERVER_ERROR = (500, "Server Error: Reference {0}")


DEPRECATED_ERROR_CODES = {}


class ApiError(Exception):
    def __init__(self, code=None, message=None, error_tuple=None, http_status_code=http.client.BAD_REQUEST, **kwargs):
        if error_tuple:
            self.error_code = error_tuple[0]
            self.error_message = error_tuple[1]
        else:
            if not (code and message):
                raise Exception('Must specify either clienterror tuple or BOTH code and message')
            self.error_code = code
            self.error_message = message

        self.status_code = http_status_code
        self.extra_fields = kwargs

    def as_dict(self):
        result = {
            'error_code': self.error_code,
            'error_message': self.error_message
        }

        if self.extra_fields:
            result.update(self.extra_fields)

        return result

    @property
    def error_code(self):
        try:
            return self.__dict__['error_code']
        except KeyError:
            raise AttributeError("'{}' object has no attribute '{}'".format(ApiError.__name__, 'error_code'))

    @error_code.setter
    def error_code(self, value):
        if value in DEPRECATED_ERROR_CODES:
            warnings.warn("deprecated error code", DeprecationWarning)
            raise ValueError("ApiError cannot use a deprecated error code")
        self.__dict__['error_code'] = value


class ClientError(ApiError):
    def __init__(self, *args, **kwargs):
        self.save_changes = kwargs.pop('save_changes', False)
        super().__init__(*args, **kwargs)


class ClientErrorUnprocessableEntity(ClientError):
    def __init__(self, additional_info=None):
        error_code, error_message = (
            http.HTTPStatus.UNPROCESSABLE_ENTITY,
            http.HTTPStatus.UNPROCESSABLE_ENTITY.phrase
        )
        if additional_info:
            error_message = f'{error_message} : {additional_info}'

        super().__init__(error_code, error_message, http_status_code=error_code)


class ClientErrorNotFound(ClientError):
    def __init__(self, additional_info=None):
        error_code, error_message = http.client.NOT_FOUND, http.client.responses.get(http.client.NOT_FOUND)
        if additional_info:
            error_message += ' : ' + additional_info
        super().__init__(error_code, error_message, http_status_code=http.client.NOT_FOUND)


class ClientErrorForbidden(ClientError):
    def __init__(self, additional_info=None, **kwargs):
        error_code, error_message = FORBIDDEN
        if additional_info:
            error_message += ": %s" % additional_info
        super().__init__(
            error_code, error_message, http_status_code=http.client.FORBIDDEN, **kwargs
        )


class ClientErrorUnauthorized(ClientError):
    def __init__(self, additional_info=None, **kwargs):
        error_code, error_message = AUTHORIZATION_FAILURE
        if additional_info:
            error_message = f'{error_message} : {additional_info}'

        super().__init__(
            error_code, error_message, http_status_code=http.HTTPStatus.UNAUTHORIZED, **kwargs
        )


class ClientErrorExternalServiceFailure(ClientError):
    def __init__(self, additional_info=None):
        error_code, error_message = EXTERNAL_REQUEST_FAILURE
        if additional_info:
            error_message += ": {0}".format(additional_info)
        super().__init__(error_code, error_message)


class ClientErrorRequestThrottled(ClientError):
    def __init__(self):
        error_code, error_message = REQUEST_THROTTLED
        super().__init__(error_code, error_message, http_status_code=429)


class ClientErrorTimedOut(ClientError):
    def __init__(self, additional_info=None):
        error_code, error_message = TIMED_OUT
        if additional_info:
            error_message += ": %s" % additional_info
        super().__init__(error_code, error_message, http_status_code=http.client.REQUEST_TIMEOUT)


class ClientErrorResponseWrapper(ClientError):
    def __init__(self, response):
        error_code = response.status_code
        error_message = response.content
        status_code = response.status_code
        super().__init__(error_code, error_message, http_status_code=status_code)


class ClientErrorExtraFields(ClientError):
    def __init__(self, extra_fields=None):
        error_code, error_message = EXTRA_FIELDS
        if extra_fields:
            error_message += ": %s" % ', '.join(extra_fields)
        super().__init__(error_code, error_message)


class ClientErrorReadOnlyFields(ClientError):
    def __init__(self, read_only_fields=None):
        error_code, error_message = READ_ONLY_FIELDS
        if read_only_fields:
            error_message += ": %s" % ', '.join(read_only_fields)
        super().__init__(error_code, error_message)


class ClientErrorMissingFields(ClientError):
    def __init__(self, missing_fields=None, extra_message=None):
        error_code, error_message = MISSING_FIELDS
        if missing_fields:
            error_message += ": %s" % ', '.join(missing_fields)
        if extra_message:
            error_message += ": {0}".format(extra_message)
        super().__init__(error_code, error_message)


class ClientErrorInvalidFieldValues(ClientError):
    def __init__(self, invalid_fields=None, extra_message=None):
        error_code, error_message = INVALID_FIELD_VALUES
        if invalid_fields:
            error_message += ": %s" % ', '.join(invalid_fields)
        if extra_message:
            error_message += ": {0}".format(extra_message)
        super().__init__(error_code, error_message)


class ServerError(ApiError):
    def __init__(self):
        error_code, error_message = LOGGED_SERVER_ERROR
        error_message = error_message.format(uuid.uuid4())
        logging.exception(error_message)

        self._cause = None

        super().__init__(error_code, error_message, http_status_code=http.client.INTERNAL_SERVER_ERROR)

    @property
    def __cause__(self):
        return self._cause

    @__cause__.setter
    def __cause__(self, value):
        """Augment __cause__ setter to capture the relevant __traceback__ for the cause (inspired by Python 3)"""
        self._cause = value
        _, cause, traceback = sys.exc_info()
        if value is cause and not hasattr(cause, '__traceback__'):
            value.__traceback__ = traceback



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import django.test
import mock

from django_declarative_apis.machinery import filtering

from . import filters
from . import models


class FiltersTestCase(django.test.TestCase):
    def setUp(self):
        self.test_model = models.TestModel.objects.create(id=1, int_field=1)
        self.p1 = models.ParentModel.objects.create(nonstandard_id=2, name='p1')
        self.p1c1 = models.ChildModel.objects.create(id=3,
                                                     test=self.test_model,
                                                     name='p1c1',
                                                     parent=self.p1)
        self.p1c2 = models.ChildModel.objects.create(id=4,
                                                     test=self.test_model,
                                                     name='p1c2',
                                                     parent=self.p1)
        self.p1.favorite = self.p1c1
        self.p1.save()

        self.root = models.RootNode.objects.create(id=5,
                                                   parent_field=self.p1)
        self.p1.root = self.root
        self.p1.save()

    def test_expandable_field_not_expanded_by_default(self):
        filtered = filtering.apply_filters_to_object(self.root, filters.DEFAULT_FILTERS)
        self.assertEqual(4, len(filtered))
        self.assertTrue('parent_field' in filtered)
        self.assertEqual(1, len(filtered['parent_field']))
        self.assertEqual(self.p1.nonstandard_id, filtered['parent_field']['nonstandard_id'])

    def test_expand_single_level(self):
        filtered = filtering.apply_filters_to_object(self.root,
                                                     filters.DEFAULT_FILTERS,
                                                     expand_header='parent_field')
        self.assertEqual(4, len(filtered))
        self.assertTrue('parent_field' in filtered)
        self.assertEqual(5, len(filtered['parent_field']))
        self.assertEqual(self.p1.nonstandard_id, filtered['parent_field']['nonstandard_id'])
        self.assertEqual(self.p1.name, filtered['parent_field']['name'])
        self.assertEqual(self.p1c1.name, filtered['parent_field']['favorite']['name'])
        self.assertEqual(2, len(filtered['parent_field']['children']))
        for child in filtered['parent_field']['children']:
            self.assertEqual(1, len(child))
            self.assertTrue('name' in child)

    def test_expand_multi_level(self):
        filtered = filtering.apply_filters_to_object(self.root,
                                                     filters.DEFAULT_FILTERS,
                                                     expand_header='parent_field.favorite')
        self.assertEqual(4, len(filtered))
        self.assertTrue('parent_field' in filtered)
        self.assertEqual(5, len(filtered['parent_field']))
        self.assertEqual(self.p1.nonstandard_id, filtered['parent_field']['nonstandard_id'])
        self.assertEqual(self.p1.name, filtered['parent_field']['name'])
        self.assertEqual(5, len(filtered['parent_field']['favorite']))
        self.assertEqual(self.p1c1.name, filtered['parent_field']['favorite']['name'])
        self.assertEqual(2, len(filtered['parent_field']['children']))
        for child in filtered['parent_field']['children']:
            self.assertEqual(1, len(child))
            self.assertTrue('name' in child)

    def test_expand_multi_level_reverse_fk_relation(self):
        filtered = filtering.apply_filters_to_object(self.root,
                                                     filters.DEFAULT_FILTERS,
                                                     expand_header='parent_field.children')
        self.assertEqual(4, len(filtered))
        self.assertTrue('parent_field' in filtered)
        self.assertEqual(5, len(filtered['parent_field']))
        self.assertEqual(self.p1.nonstandard_id, filtered['parent_field']['nonstandard_id'])
        self.assertEqual(self.p1.name, filtered['parent_field']['name'])
        self.assertEqual(1, len(filtered['parent_field']['favorite']))
        self.assertTrue('name' in filtered['parent_field']['favorite'])
        self.assertEqual(self.p1c1.name, filtered['parent_field']['favorite']['name'])
        self.assertEqual(2, len(filtered['parent_field']['children']))
        for child in filtered['parent_field']['children']:
            self.assertEqual(5, len(child))
            self.assertTrue('pk' in child)
            self.assertTrue('name' in child)
            self.assertTrue('test' in child)
            self.assertTrue('parent' in child)
            self.assertEqual(1, len(child['parent']))

    def test_expand_multi_level_more_than_one_field(self):
        filtered = filtering.apply_filters_to_object(self.root,
                                                     filters.DEFAULT_FILTERS,
                                                     expand_header='parent_field,parent_field.children, parent_field.favorite.test')
        self.assertEqual(4, len(filtered))
        self.assertTrue('parent_field' in filtered)
        self.assertEqual(5, len(filtered['parent_field']))
        self.assertEqual(self.p1.nonstandard_id, filtered['parent_field']['nonstandard_id'])
        self.assertEqual(self.p1.name, filtered['parent_field']['name'])
        self.assertEqual(5, len(filtered['parent_field']['favorite']))
        self.assertTrue('pk' in filtered['parent_field']['favorite'])
        self.assertTrue('name' in filtered['parent_field']['favorite'])
        self.assertTrue('test' in filtered['parent_field']['favorite'])
        self.assertEqual(3, len(filtered['parent_field']['favorite']['test']))
        self.assertTrue('parent' in filtered['parent_field']['favorite'])
        self.assertEqual(self.p1c1.name, filtered['parent_field']['favorite']['name'])
        self.assertEqual(2, len(filtered['parent_field']['children']))
        for child in filtered['parent_field']['children']:
            self.assertEqual(5, len(child))
            self.assertTrue('pk' in child)
            self.assertTrue('name' in child)
            self.assertTrue('test' in child)
            self.assertEqual(1, len(child['test']))
            self.assertTrue('parent' in child)
            self.assertEqual(1, len(child['parent']))

    def test_expandable_properties(self):
        filtered = filtering.apply_filters_to_object(self.test_model,
                                                     filters.DEFAULT_FILTERS,
                                                     expand_header='expandable_dict,expandable_string')

        self.assertEqual(5, len(filtered))
        self.assertTrue('expandable_dict' in filtered)
        self.assertEqual(filtered['expandable_dict'], models.TestModel.EXPANDABLE_DICT_RETURN)
        self.assertTrue('expandable_string' in filtered)
        self.assertEqual(filtered['expandable_string'], models.TestModel.EXPANDABLE_STRING_RETURN)
        self.assertTrue('__expandable__' in filtered)
        self.assertTrue('expandable_dict' in filtered['__expandable__'])
        self.assertTrue('expandable_string' in filtered['__expandable__'])

        with mock.patch('tests.models.TestModel.expandable_dict') as dict_mock, \
                mock.patch('tests.models.TestModel.expandable_string') as str_mock:
            filtered = filtering.apply_filters_to_object(self.test_model,
                                                         filters.DEFAULT_FILTERS)

            self.assertEqual(3, len(filtered))
            self.assertFalse('expandable_dict' in filtered)
            dict_mock.assert_not_called()
            self.assertFalse('expandable_string' in filtered)
            str_mock.assert_not_called()
            self.assertTrue('__expandable__' in filtered)
            self.assertTrue('expandable_dict' in filtered['__expandable__'])
            self.assertTrue('expandable_string' in filtered['__expandable__'])

    def test_expandable_absent_if_no_expandable_fields(self):
        filtered = filtering.apply_filters_to_object(self.test_model,
                                                     filters.DEFAULT_FILTERS_NO_EXPANDABLE)

        self.assertEqual(4, len(filtered))
        self.assertTrue('expandable_dict' in filtered)
        self.assertEqual(filtered['expandable_dict'], models.TestModel.EXPANDABLE_DICT_RETURN)
        self.assertTrue('expandable_string' in filtered)
        self.assertEqual(filtered['expandable_string'], models.TestModel.EXPANDABLE_STRING_RETURN)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import logging
from pydoc import locate
from celery.task import task as celery_task
import celery
import time
from typing import NamedTuple
import kombu.exceptions

from django.conf import settings
from django.core.cache import cache
import django.db.models

try:
    from newrelic import agent as newrelic_agent
except ImportError:
    newrelic_agent = None

try:
    import cid.locals
    _get_correlation_id = cid.locals.get_cid
except ImportError:
    _get_correlation_id = lambda: None

JOB_COUNT_CACHE_KEY = 'future_task_runner:job_id'
QUEUE_LENGTH_CACHE_KEY = 'future_task_runner:current_queue_length'

process_task_count = 0
logger = logging.getLogger(__name__)


def get_current_queue_length():
    return cache.get(QUEUE_LENGTH_CACHE_KEY, 0)


def _get_task_job_count():
    try:
        return cache.incr(JOB_COUNT_CACHE_KEY)
    except ValueError:
        cache.set(JOB_COUNT_CACHE_KEY, 0)
        return 0


def _log_task_stats(method_name, resource_instance_id, scheduled_execution_delay, task_creation_time,
                    task_job_count, correlation_id=None):
    global process_task_count
    process_task_count += 1
    if task_creation_time:
        now = time.time()
        scheduled_execution_time = task_creation_time + scheduled_execution_delay
        current_job_count = cache.get(JOB_COUNT_CACHE_KEY, 0)
        wait_time = now - task_creation_time
        queue_delay = now - scheduled_execution_time
        queue_length = current_job_count - task_job_count
        delivery_info = celery.current_task.request.delivery_info
        queue = delivery_info.get('exchange') or 'default'
        routing_key = delivery_info.get('routing_key') or 'default'

        cache.set(QUEUE_LENGTH_CACHE_KEY, queue_length)

        if newrelic_agent:
            newrelic_agent.record_custom_event('task_runner:queue_length',
                                               {'queue_length': queue_length,
                                                'queue': queue,
                                                'routing_key': routing_key,
                                                'wait_time_seconds': wait_time,
                                                'queue_delay_seconds': queue_delay,
                                                'process_task_count': process_task_count,
                                                'correlation_id': correlation_id,})

        logger.info(
            'method=%s, resource_id=%s, queue_length=%s, queue=%s, routing_key=%s, task_wait_time=%s, '
            'task_queue_delay=%s, process_task_count=%s, correlation_id=%s',
            method_name, resource_instance_id,
            queue_length,
            queue,
            routing_key,
            wait_time,
            queue_delay,
            process_task_count,
            correlation_id
        )
    else:
        logger.info('method=%s, resource_id=%s, correlation_id=%s', method_name, resource_instance_id, correlation_id)


def _log_retry_stats(method_name, resource_instance_id, correlation_id):
    if newrelic_agent:
        newrelic_agent.record_custom_event('task_runner:retry',
                                           {'method_name': method_name,
                                            'resource_instance_id': resource_instance_id})
        logger.warn(
            'will retry task: method=%s, resource_id=%s, correlation_id=%s', method_name, resource_instance_id,
            correlation_id
        )


class RetryParams(NamedTuple):
    retries_remaining: int
    retry_exception_filter: tuple
    queue: str
    routing_key: str
    countdown: int


@celery_task(ignore_results=True,
             time_limit=getattr(settings, 'DDA_DEFERRED_TASK_TIME_LIMIT', 999999),
             soft_time_limit=getattr(settings, 'DDA_DEFERRED_TASK_SOFT_TIME_LIMIT', 999999))
def future_task_runner(endpoint_class_name, endpoint_method_name, resource_class_name, resource_instance_id,
                       task_job_count=0, task_creation_time=None, scheduled_execution_delay=0, task_args=None,
                       retry_params=None, correlation_id=None):
    endpoint_class = locate(endpoint_class_name)
    resource_class = locate(resource_class_name)
    resource_instance = resource_class.objects.get(pk=resource_instance_id)
    endpoint_task = getattr(endpoint_class, endpoint_method_name)

    _log_task_stats(endpoint_method_name, resource_instance_id, scheduled_execution_delay, task_creation_time,
                    task_job_count, correlation_id)

    logger.info('future_task_runner: method=%s, resource_id=%s', endpoint_method_name, resource_instance_id)

    if task_args:
        args, kwargs = task_args
    else:
        args, kwargs = [], {}

    try:
        endpoint_task.task_runner(resource_instance, *args, **kwargs)
    except Exception as e:
        if retry_params is None:
            raise

        retry_params = RetryParams(*retry_params)
        if retry_params.retry_exception_filter\
                and not any([
                    f'{e.__class__.__module__}.{e.__class__.__name__}' == ex_type
                    for ex_type in retry_params.retry_exception_filter
                ]):
            # this exception is not retryable
            raise

        if retry_params.retries_remaining == 0:
            # out of chances
            raise

        _log_retry_stats(endpoint_method_name, resource_instance_id, correlation_id)
        task_runner_args = (endpoint_class_name, endpoint_method_name, resource_class_name, resource_instance_id)
        task_runner_kwargs = {
            'task_creation_time': time.time(),
            'scheduled_execution_delay': scheduled_execution_delay,
            'task_args': (args, kwargs),
        }
        schedule_future_task_runner(
            task_runner_args,
            task_runner_kwargs,
            retries=retry_params.retries_remaining-1,
            countdown=retry_params.countdown * 2 or 1,
            queue=retry_params.queue,
            routing_key=retry_params.routing_key,
        )


def schedule_future_task_runner(task_runner_args, task_runner_kwargs,
                                retries=0,
                                retry_exception_filter=(),
                                queue=None,
                                routing_key=None,
                                countdown=0,
                                delay=0):
    task_runner_kwargs['task_job_count'] = _get_task_job_count()
    task_runner_kwargs['retry_params'] = RetryParams(
        retries_remaining=retries,
        retry_exception_filter=tuple(
            f'{exc.__module__}.{exc.__name__}' for exc in retry_exception_filter
        ),
        queue=queue,
        routing_key=routing_key,
        countdown=countdown
    )
    task_runner_kwargs['correlation_id'] = _get_correlation_id()

    if getattr(settings, 'DECLARATIVE_ENDPOINT_TASKS_FORCE_SYNCHRONOUS', False):
        logger.info('Processing tasks synchronously')
        future_task_runner.apply(task_runner_args, task_runner_kwargs)
    else:
        MAX_ATTEMPTS = 3
        for attempt in range(MAX_ATTEMPTS):
            # XXX: This is an attempt to skirt around an unsolved, low repro issue somewhere in the celery/kombu/redis-py stack.
            # Once in a while, a connection in the pool will timeout prior to a health check being called in redis-py and
            # will result in an error being raised here. This should be removed once the issue has been sorted out.
            # Note: This is around the use of redis-py in celery where celery's event loop is not running
            # https://github.com/celery/kombu/issues/1019
            try:
                future_task_runner.apply_async(task_runner_args, task_runner_kwargs, queue=queue, routing_key=routing_key, countdown=countdown+delay)
                return
            except kombu.exceptions.OperationalError as err:
                logger.warn('kombu.exceptions.OperationalError (attempt: %s)', attempt)
                if attempt >= MAX_ATTEMPTS - 1:
                    if getattr(settings, 'DECLARATIVE_ENDPOINT_TASKS_SYNCHRONOUS_FALLBACK'):
                        logger.warn('Falling back to executing task synchronously')
                        future_task_runner.apply(task_runner_args, task_runner_kwargs)
                        return
                    raise err


@celery_task(ignore_results=True,
             time_limit=getattr(settings, 'DDA_DEFERRED_TASK_TIME_LIMIT', 999999),
             soft_time_limit=getattr(settings, 'DDA_DEFERRED_TASK_SOFT_TIME_LIMIT', 999999))
def resource_task_runner(resource_class_name, resource_method_name, resource_instance_id,
                         task_job_count=0, task_creation_time=None, scheduled_execution_delay=0, task_args=None,
                         correlation_id=None):
    resource_class = locate(resource_class_name)
    resource_instance = resource_class.objects.get(pk=resource_instance_id)
    resource_method = getattr(resource_instance, resource_method_name)

    _log_task_stats("{0}.{1}".format(resource_class_name, resource_method_name),
                    resource_instance_id,
                    scheduled_execution_delay,
                    task_creation_time,
                    task_job_count)
    if task_args:
        args, kwargs = task_args
    else:
        args, kwargs = [], {}

    resource_method(*args, **kwargs)


def schedule_resource_task_runner(resource_bound_method,
                                  task_args=None,
                                  task_kwargs=None,
                                  queue=None,
                                  routing_key=None,
                                  delay=0):
    resource = resource_bound_method.__self__
    assert isinstance(resource, django.db.models.Model), \
        "resource must be an instance of django.db.models.Model to run as deferred task"

    resource_class_name = '{0}.{1}'.format(resource.__module__, resource.__class__.__name__)
    resource_id = resource.pk

    task_runner_args = (resource_class_name, resource_bound_method.__name__, str(resource_id))
    task_runner_kwargs = {
        'task_creation_time': time.time(),
        'scheduled_execution_delay': delay,
        'task_args': (task_args or (), task_kwargs or {}),
        'task_job_count': _get_task_job_count(),
        'correlation_id': _get_correlation_id(),
    }

    resource_task_runner.apply_async(task_runner_args,
                                     task_runner_kwargs,
                                     queue=queue,
                                     routing_key=routing_key)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import abc
import collections
import logging
import random
import string
import time

from django.db import models as django_models

from . import errors
from . import tasks


class EndpointAttribute(metaclass=abc.ABCMeta):
    _next_attribute_number = 0

    @classmethod
    def claim_attribute_number(cls):
        next_serial_number = cls._next_attribute_number
        cls._next_attribute_number += 1
        return next_serial_number

    def __init__(self, name=None, required=False, description=None, hidden=False, advanced=False):
        super(EndpointAttribute, self).__init__()

        self.hidden = hidden
        self.name = name  # will be populated by EndpointDefinitionMeta
        self.advanced = advanced
        self.required = required
        self.description = description
        self.attribute_number = EndpointAttribute.claim_attribute_number()

    def __get__(self, owner_instance, owner_class):
        if not owner_instance:
            # If accessed from the class, return the EndpointAttribute itself
            return self

        if not self.name:
            raise ValueError("All EndpointAttribute objects must have a name before they can be accessed "
                             "from an instance")

        # The instance dictionary is used as a cache
        if self.name in owner_instance.__dict__:
            return owner_instance.__dict__[self.name]

        # Delegate to _get_value_for_instance and cache the result within the instance's dict
        value = self.get_instance_value(owner_instance, owner_class)
        owner_instance.__dict__[self.name] = value
        return value

    @abc.abstractmethod
    def get_instance_value(self, owner_instance, owner_class):
        raise NotImplementedError()


class RequestProperty(EndpointAttribute):
    __hidden_request_attribute_name = '_' + ''.join(random.choice(string.printable) for _ in range(10))

    @classmethod
    def bind_request_to_instance(cls, instance, request):
        # Hide the request within the instance
        setattr(instance, cls.__hidden_request_attribute_name, request)

    @classmethod
    def request_has_been_bound(cls, instance):
        return cls.__hidden_request_attribute_name in instance.__dict__

    def __init__(self, property_getter, **kwargs):
        super(RequestProperty, self).__init__(**kwargs)

        self.property_getter = property_getter

        # Capture the hidden attribute name within each RequestProperty instance, so lookups work through being pickled
        self.__hidden_request_attribute_name = RequestProperty.__hidden_request_attribute_name

    def __extract_request_from_instance(self, instance):
        return getattr(instance, self.__hidden_request_attribute_name, None)

    def get_instance_value(self, owner_instance, owner_class):
        request = self.__extract_request_from_instance(owner_instance)
        if not request:
            raise ValueError("A request must be bound with the instance before accessing this property")

        return self.property_getter(owner_instance, request)

    @property
    def documentation(self):
        result = {
            'name': self.name,
        }
        if self.description is not None:
            result['description'] = self.description
        return result


class TypedEndpointAttributeMixin(object):
    def __init__(self, *args, **kwargs):
        self.field_type = kwargs.pop('type', str)
        if self.field_type not in RequestField.VALID_FIELD_TYPES:
            raise NotImplementedError('Request fields of type {0} not supported'.format(self.field_type.__name__))
        super(TypedEndpointAttributeMixin, self).__init__(*args, **kwargs)

    def coerce_value_to_type(self, raw_value):
        try:
            if self.field_type == bool and not isinstance(raw_value, self.field_type):
                return 'rue' in raw_value
            else:
                if isinstance(raw_value, collections.Iterable) and not isinstance(raw_value, (str, dict)):
                    return list(self.field_type(r) for r in raw_value)
                else:
                    return self.field_type(raw_value)
        except Exception as e:
            raise errors.ClientErrorInvalidFieldValues(
                [self.name],
                'Could not parse {val} as type {type}'.format(val=raw_value, type=self.field_type.__name__))


class RequestUrlField(TypedEndpointAttributeMixin, EndpointAttribute):
    def __init__(self, *args, **kwargs):
        self.api_name = kwargs.pop('name', None)
        self.value = None
        super(RequestUrlField, self).__init__(*args, **kwargs)

    def set_value(self, value):
        self.value = value

    def get_instance_value(self, owner_instance, owner_class):
        return self.coerce_value_to_type(self.value)


class RequestAdhocQuerySet(RequestUrlField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, type=dict, **kwargs)
        self.value = {}


class RequestField(TypedEndpointAttributeMixin, RequestProperty):
    VALID_FIELD_TYPES = (bool, int, float, complex, str, dict)

    def __init__(self, *args, **kwargs):
        self.default_value = kwargs.pop('default', None)
        self.api_name = kwargs.pop('name', None)
        self.multivalued = kwargs.pop('multivalued', False)
        super(RequestField, self).__init__(property_getter=self.get_field, **kwargs)
        self.post_processor = None

    def __call__(self, post_processor):
        self.post_processor = post_processor
        return self

    @property
    def documentation(self):
        result = super(RequestField, self).documentation
        result['type'] = self.field_type
        result['multivalued'] = self.multivalued
        if self.api_name:
            result['name'] = self.api_name
        if self.default_value is not None:
            result['default_value'] = self.default_value
        return result

    def get_without_default(self, owner_instance, request):
        if request.method == 'POST':
            query_dict = request.POST
        else:
            query_dict = request.GET

        if (self.api_name or self.name) in query_dict:
            if not self.multivalued:
                raw_value = query_dict.get(self.api_name or self.name)
            else:
                raw_value = query_dict.getlist(self.api_name or self.name)
            typed_value = self.coerce_value_to_type(raw_value)
        else:
            typed_value = None

        if self.post_processor:
            return self.post_processor(owner_instance, typed_value)
        else:
            return typed_value

    def get_field(self, owner_instance, request):
        raw_value = self.get_without_default(owner_instance, request)
        if raw_value is not None:
            return raw_value
        else:
            return self.default_value


class ResourceField(RequestField):
    pass


class RequestAttribute(RequestProperty):
    def __init__(self, attribute_getter=None, required=True, default=None, **kwargs):
        super(RequestAttribute, self).__init__(property_getter=self.get_request_attribute,
                                               required=required, **kwargs)

        self.attribute_getter = attribute_getter
        self.default = default

    # Facilitated use as a decorator with arguments
    def __call__(self, attribute_getter):
        self.attribute_getter = attribute_getter
        return self

    def get_without_default(self, owner_instance, request):
        if self.attribute_getter:
            result = self.attribute_getter(owner_instance, request)
        else:
            result = getattr(request, self.name, None)
        return result

    def get_request_attribute(self, owner_instance, request):
        result = self.get_without_default(owner_instance, request)
        if result is not None:
            return result
        else:
            return self.default


class ConsumerAttribute(RequestAttribute):
    def __init__(self, *args, field_name=None, **kwargs):
        self.field_name = field_name
        super(ConsumerAttribute, self).__init__(*args, **kwargs)

    def get_without_default(self, owner_instance, request):
        consumer = request.consumer
        if self.attribute_getter:
            return self.attribute_getter(owner_instance, consumer)
        else:
            return getattr(consumer, self.field_name or self.name, None)


class RawRequestObjectProperty(RequestAttribute):
    class SafeRequestWrapper(object):
        __hidden_request_attribute_name = '_' + ''.join(random.choice(string.printable) for _ in range(10))
        __permitted_request_properties = ('build_absolute_uri', 'method', 'META')

        def __init__(self, request, additional_safe_fields=()):
            setattr(self, self.__hidden_request_attribute_name, request)
            self.additional_safe_fields = additional_safe_fields

        @property
        def body_field_names(self):
            return set(getattr(self, self.__hidden_request_attribute_name).POST.keys())

        def __getattr__(self, name):
            if (name in RawRequestObjectProperty.SafeRequestWrapper.__permitted_request_properties
                    or name in self.additional_safe_fields):
                hidden_request = getattr(self, self.__hidden_request_attribute_name)
                return getattr(hidden_request, name)
            else:
                raise AttributeError()

    def __init__(self, *args, additional_safe_fields=(), **kwargs):
        super().__init__(*args, **kwargs)
        self.additional_safe_fields = additional_safe_fields

    def get_without_default(self, owner_instance, request):
        return RawRequestObjectProperty.SafeRequestWrapper(request, additional_safe_fields=self.additional_safe_fields)


class EndpointTask(EndpointAttribute):
    STATE_NOT_RUN = 0
    STATE_RUNNING = 1
    STATE_COMPLETED = 2

    def __init__(self,
                 task_runner=None,
                 depends_on=None,  # Reference to another task that should be run before this one.  Overrides priority
                 priority=0,       # lower priority gets executed first
                 **kwargs):
        super(EndpointTask, self).__init__(**kwargs)
        self.task_runner = task_runner
        self.depends_on = depends_on
        self.task_state = EndpointTask.STATE_NOT_RUN
        self.priority = priority

    def __call__(self, task_runner):
        self.task_runner = task_runner
        return self

    def _run_task(self, owner_instance):
        self.task_runner(owner_instance)

    def run(self, owner_instance):
        assert self.task_state != EndpointTask.STATE_RUNNING, "Circular task reference detected!"
        try:
            self.task_state = EndpointTask.STATE_RUNNING

            depends_on = self.depends_on
            if isinstance(depends_on, str):
                depends_on = getattr(owner_instance, depends_on)

            if depends_on and (depends_on.task_state != EndpointTask.STATE_COMPLETED):
                assert not isinstance(depends_on, DeferrableEndpointTask),\
                    "DeferredEndpointTask cannot be used as depends_on arg"
                depends_on.run(owner_instance)

            self._run_task(owner_instance)

        finally:
            self.task_state = EndpointTask.STATE_COMPLETED

    def get_instance_value(self, owner_instance, owner_class):
        return self


class DeferrableEndpointTask(EndpointTask):
    @staticmethod
    def unwrap_staticmethod(method):
        assert isinstance(method,
                          staticmethod), "Deferrable task methods MUST be staticmethods.  Hint: the staticmethod " \
                                         "decorator should come after the deferrable task decorator "
        # we're effectively unwrapping the staticmethod here...
        return method.__func__

    def __init__(self,
                 task_runner=None,
                 delay=None,            # delay in seconds before running the task.  Requires deferred=True
                 always_defer=True,     # True: run task in deferred queue even when delay=0
                 task_args_factory=None,
                 queue=None,
                 routing_key=None,
                 retries=0,
                 retry_exception_filter=(),
                 **kwargs):
        super(DeferrableEndpointTask, self).__init__(**kwargs)
        if task_runner:
            self.task_runner = DeferrableEndpointTask.unwrap_staticmethod(task_runner)
        else:
            self.task_runner = None

        self.delay = delay
        self.always_defer = always_defer
        self.queue = queue
        self.routing_key = routing_key
        self.retries = retries
        self.retry_exception_filter = retry_exception_filter

        assert task_args_factory is None or callable(task_args_factory)
        self.task_args_factory = task_args_factory

    def __call__(self, task_runner):
        self.task_runner = DeferrableEndpointTask.unwrap_staticmethod(task_runner)
        return self

    def _resolve_maybe_callable(self, owner_instance, maybe_callable):
        if callable(maybe_callable):
            return maybe_callable(owner_instance)
        else:
            return maybe_callable

    def _run_task(self, owner_instance):
        resource = owner_instance.resource

        assert isinstance(resource, django_models.Model),\
            "resource must be an instance of django.db.models.Model to run as deferred task"

        delay = self._resolve_maybe_callable(owner_instance, self.delay) or 0
        always_defer = self._resolve_maybe_callable(owner_instance, self.always_defer)

        if self.task_args_factory:
            task_args, task_kwargs = self.task_args_factory(owner_instance)
        else:
            task_args, task_kwargs = [], {}

        if delay == 0 and not always_defer:
            self.task_runner(*([resource] + list(task_args)), **task_kwargs)
        else:
            if resource.pk is None:
                resource.save()
            resource_id = resource.pk
            resource_class_name = '{0}.{1}'.format(resource.__module__, resource.__class__.__name__)
            endpoint_class_name = '{0}.{1}'.format(owner_instance.__module__, owner_instance.__class__.__name__)
            task_runner_args = (endpoint_class_name, self.task_runner.__name__, resource_class_name, str(resource_id))
            task_runner_kwargs = {
                'task_creation_time': time.time(),
                'scheduled_execution_delay': delay,
                'task_args': (task_args, task_kwargs),
            }
            tasks.schedule_future_task_runner(
                task_runner_args,
                task_runner_kwargs,
                retries=self.retries,
                retry_exception_filter=self.retry_exception_filter,
                delay=delay,
                queue=self.queue,
                routing_key=self.routing_key,
            )


class RequestFieldGroup(RequestProperty):
    def __init__(self, *component_field_getters, **kwargs):
        super(RequestFieldGroup, self).__init__(property_getter=self.get_value,
                                                **kwargs)
        self.component_field_getters = component_field_getters
        self.component_field_names = []
        for component_field_getter in self.component_field_getters:
            component_field_getter.required = False

    def __call__(self, *component_field_getters):
        self.component_field_getters = component_field_getters
        return self

    def _get_request_dict(self, request):
        if request.method == 'GET':
            return request.GET
        elif request.method == 'POST':
            return request.POST
        else:
            return {}

    def _get_missing_component_fields(self, owner_instance, request):
        self.component_field_names = map(lambda x: x.name, self.component_field_getters)
        request_dict = self._get_request_dict(request)
        missing_fields = []
        for getter in self.component_field_getters:
            result = getter.get_without_default(owner_instance, request)
            if result is None:
                missing_fields.append(getter.name)
        return missing_fields


class RequireOneAttribute(RequestFieldGroup):
    def get_value(self, owner_instance, request):
        missing_fields = self._get_missing_component_fields(owner_instance, request)

        # all but one field should be missing from the request
        if len(self.component_field_getters) - len(missing_fields) == 1:
            return True
        else:
            raise errors.ClientErrorMissingFields(self.component_field_names,
                                                     extra_message="Exactly one field must be populated")


class RequireAllAttribute(RequestFieldGroup):
    def get_value(self, owner_instance, request):
        missing_fields = self._get_missing_component_fields(owner_instance, request)

        # no fields should be missing from the request
        if len(missing_fields) == 0:
            return True
        else:
            raise errors.ClientErrorMissingFields(self.component_field_names,
                                                     extra_message="All fields must be populated")


class RequireAllIfAnyAttribute(RequestFieldGroup):
    def get_value(self, owner_instance, request):
        missing_fields = self._get_missing_component_fields(owner_instance, request)

        # either all present or all missing
        if (len(missing_fields) == 0) or (len(missing_fields) == len(self.component_field_getters)):
            return True
        else:
            raise errors.ClientErrorMissingFields(self.component_field_names,
                                                     extra_message="All fields must be populated")


class Aggregate(EndpointAttribute):
    def __init__(self, aggregation_function=None, **kwargs):
        self.aggregation_function = aggregation_function
        self.depends_on = kwargs.pop('depends_on', None)
        super(Aggregate, self).__init__(**kwargs)

    def __call__(self, aggregation_function):
        self.aggregation_function = aggregation_function
        return self

    def get_instance_value(self, owner_instance, owner_class):
        if not RequestProperty.request_has_been_bound(owner_instance):
            raise Exception('Request must be bound to endpoint before accessing aggregate values')

        return self.aggregation_function(owner_instance)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import json
from http import HTTPStatus

from django.test import TestCase

from . import testutils

from django_declarative_apis import models as auth_models

class DeclarativeApisTestCase(TestCase):
    client_class = testutils.DeclarativeApisOAuthClient

    def setUp(self):
        self.consumer = auth_models.OauthConsumer.objects.create()

    def test_simplest_endpoint(self):
        response = self.client.get('/simple',
                                   consumer=self.consumer,
                                   expected_status_code=HTTPStatus.OK)

    def test_typed_parameter(self):
        response = self.client.get('/simple?int_type_field=not_an_int',
                                   consumer=self.consumer,
                                   expected_status_code=HTTPStatus.BAD_REQUEST)
        error = json.loads(response.content.decode('utf-8'))
        self.assertEqual(error['error_code'], 703)
        self.assertTrue('Invalid values for field(s): int_type_field')



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import json
import unittest
import mock
import http
import urllib.parse

import kombu.exceptions
import django.core.exceptions
from django.db import models
from django.core.cache import cache
import django.test

from django_declarative_apis import machinery, models as dda_models
from django_declarative_apis.machinery import errors, filtering, tasks
from django_declarative_apis.machinery.tasks import future_task_runner
from django_declarative_apis.resources.utils import HttpStatusCode, rc

import tests.models
from tests import testutils

_TEST_RESOURCE = {'foo': 'bar'}


class EndpointResourceAttributeTestCase(testutils.RequestCreatorMixin, unittest.TestCase):
    def test_call(self):
        attrib = machinery.EndpointResourceAttribute(str)
        self.assertEqual(attrib(lambda: 'foo'), attrib)

    def test_get_instance_value(self):
        attrib = machinery.EndpointResourceAttribute(str)
        self.assertEqual(attrib.get_instance_value(None, None), attrib)

    def test_get_instance_value_dict(self):
        expected_resource = {'foo': 'bar'}

        class _TestEndpoint(machinery.EndpointDefinition):
            @machinery.endpoint_resource(type=dict)
            def resource(self):
                return expected_resource

        endpoint = _TestEndpoint()
        self.assertEqual(endpoint.resource, expected_resource)

    def test_get_instance_value_custom_type(self):
        class _TestResource:
            pass

        expected_resource = _TestResource()

        class _TestEndpoint(machinery.EndpointDefinition):
            @machinery.endpoint_resource(type=_TestResource)
            def resource(self):
                return expected_resource

        endpoint = _TestEndpoint()
        res = endpoint.resource
        self.assertIsInstance(res, _TestResource)
        # attribute should be set in get_instance_value
        self.assertEqual(res._api_filter, None)

    def test_get_instance_value_instance_not_found(self):
        class _TestErrorEndpoint(machinery.EndpointDefinition):
            @machinery.endpoint_resource(type=dict)
            def resource(self):
                raise django.core.exceptions.ObjectDoesNotExist

        endpoint = _TestErrorEndpoint()

        try:
            endpoint.resource
            self.fail('This should have failed')
        except errors.ClientErrorNotFound as err:
            self.assertEqual(err.error_code, http.HTTPStatus.NOT_FOUND)
            self.assertEqual(err.error_message, 'Not Found : dict instance not found')
            self.assertEqual(err.status_code, http.HTTPStatus.NOT_FOUND)
            self.assertFalse(err.save_changes)
            self.assertEqual(err.extra_fields, {})


class EndpointResponseAttributeTestCase(testutils.RequestCreatorMixin, unittest.TestCase):
    def test_call(self):
        attr = machinery.EndpointResponseAttribute(str)
        self.assertEqual(attr.get_instance_value(None, None), attr)

    def test_instance_value(self):
        expected_response = {'foo': 'bar'}
        class _TestEndpoint(machinery.EndpointDefinition):
            @machinery.endpoint_response(type=dict)
            def response(self):
                return expected_response

        endpoint = _TestEndpoint()
        self.assertEqual(endpoint.response, expected_response)

    def test_instance_value_with_filter(self):
        class _TestResource:
            def __init__(self, name, secret):
                self.name = name
                self.secret = secret

        expected_response = _TestResource(name='foo', secret='bar')
        obj_filter = {'name': filtering.ALWAYS, 'secret': filtering.NEVER}
        class _TestEndpoint(machinery.EndpointDefinition):
            @machinery.endpoint_response(type=dict, filter=obj_filter)
            def response(self):
                return expected_response

        endpoint = _TestEndpoint()
        resp = endpoint.response

        self.assertIsInstance(resp, _TestResource)
        self.assertEqual(resp._api_filter, obj_filter)


class EndpointBinderTestCase(django.test.TestCase):

    @mock.patch('django_declarative_apis.machinery.logging')
    def test_get_response_with_non_client_error(self, mock_logging):
        class _TestException(Exception):
            pass

        class _TestEndpoint(machinery.EndpointDefinition):
            @property
            def resource(self):
                return {}

        endpoint = _TestEndpoint()
        manager = machinery.EndpointBinder.BoundEndpointManager(
            machinery._EndpointRequestLifecycleManager(endpoint),
            endpoint
        )
        manager.binding_exc_info = _TestException, _TestException('something bad happened'), None
        self.assertRaises(_TestException, manager.get_response)
        mock_logging.error.assert_called_with("('something bad happened',)\nNone")

    def test_get_response_with_dirty_resource(self):
        class _TestResource:
            def is_dirty(self, check_relationship=False):
                return True

            def save(self):
                pass

        class _TestEndpoint(machinery.EndpointDefinition):
            @machinery.endpoint_resource(type=_TestResource)
            def resource(self):
                return _TestResource()

        endpoint = _TestEndpoint()
        manager = machinery.EndpointBinder.BoundEndpointManager(
            machinery._EndpointRequestLifecycleManager(endpoint),
            endpoint
        )

        class _FakeRequest():
            META = {}
        manager.bound_endpoint.request = _FakeRequest()

        with mock.patch.object(_TestResource, 'save', return_value=None) as mock_save:
            manager.get_response()
            # save is called before and after tasks. since we've hardcoded _TestResource.is_dirty to return True,
            # both of them should fire
            self.assertEqual(mock_save.call_count, 2)

    def test_get_response_with_client_error_while_executing_tasks(self):
        class _TestResource:
            def is_dirty(self, check_relationship=False):
                return True

            def save(self):
                pass

        class _TestEndpoint(machinery.EndpointDefinition):
            @machinery.endpoint_resource(type=_TestResource)
            def resource(self):
                return _TestResource()

            @machinery.task
            def raise_an_exception(self):
                raise errors.ClientError(
                    code=http.HTTPStatus.BAD_REQUEST,
                    message='something bad happened',
                    save_changes=error_should_save_changes
                )

        for error_should_save_changes in (True, False):
            with mock.patch.object(_TestResource, 'save') as mock_save:
                endpoint = _TestEndpoint()
                manager = machinery.EndpointBinder.BoundEndpointManager(
                    machinery._EndpointRequestLifecycleManager(endpoint),
                    endpoint
                )
                try:
                    manager.get_response()
                    self.fail('This should have failed')
                except errors.ClientError as err:
                    # save should be called twice if the exception says the resource should be saved: once before
                    # tasks are executed and once during exception handling.
                    self.assertEqual(mock_save.call_count, 2 if error_should_save_changes else 1)


    def test_get_response_custom_http_response(self):
        expected_data = {'foo': 'bar'}
        expected_response = django.http.HttpResponse(content=json.dumps(expected_data))

        class _TestEndpoint(machinery.EndpointDefinition):
            @property
            def resource(self):
                pass

            @property
            def response(self):
                return expected_response

        endpoint = _TestEndpoint()
        manager = machinery.EndpointBinder.BoundEndpointManager(
            machinery._EndpointRequestLifecycleManager(endpoint),
            endpoint
        )

        status, resp = manager.get_response()
        self.assertEqual(status, http.HTTPStatus.OK)
        self.assertEqual(resp, expected_response)
        self.assertEqual(json.loads(resp.content), expected_data)

    def test_get_response_custom_http_error(self):
        expected_data = {'error': 'something bad happened'}

        class _TestEndpoint(machinery.EndpointDefinition):
            @property
            def resource(self):
                pass

            @property
            def response(self):
                return django.http.HttpResponse(content=json.dumps(expected_data), status=http.HTTPStatus.BAD_REQUEST)

            @property
            def http_status(self):
                # TODO: it's kind of strange that http_status has to be defined if the response returns a custom
                # http response
                return http.HTTPStatus.BAD_REQUEST

        endpoint = _TestEndpoint()
        manager = machinery.EndpointBinder.BoundEndpointManager(
            machinery._EndpointRequestLifecycleManager(endpoint),
            endpoint
        )

        try:
            manager.get_response()
            self.fail('This should have failed')
        except HttpStatusCode as err:
            self.assertEqual(err.response.status_code, http.HTTPStatus.BAD_REQUEST)
            self.assertEqual(json.loads(err.response.content), expected_data)

    def test_get_response_with_filtered_list(self):
        class _TestResource:
            def __init__(self, name, secret):
                self.name = name
                self.secret = secret

        class _QuerySet(list):
            pass

        data = _QuerySet([_TestResource('foo', 'bar'), _TestResource('bar', 'baz')])

        filter_def = {
            _TestResource: {'name': filtering.ALWAYS, 'secret': filtering.NEVER},
        }
        class _TestEndpoint(machinery.EndpointDefinition):
            @machinery.endpoint_resource(type=_TestResource, filter=filter_def)
            def resource(self):
                return data

            @property
            def response(self):
                return {
                    'people': self.resource
                }

        endpoint = _TestEndpoint()
        manager = machinery.EndpointBinder.BoundEndpointManager(
            machinery._EndpointRequestLifecycleManager(endpoint),
            endpoint
        )
        status, resp = manager.get_response()
        self.assertEqual(status, http.HTTPStatus.OK)
        # make sure the list is in the expected order
        resp['people'].sort(key=lambda p: p['name'].lower())
        self.assertEqual(resp, {'people': [{'name': 'bar'}, {'name': 'foo'}]})


class EndpointFilteringTestCase(testutils.RequestCreatorMixin, django.test.TestCase):
    from django_declarative_apis.machinery.filtering import ALWAYS, NEVER, IF_TRUTHY

    class DummyClassOne(object):
        def __init__(self):
            self.foo = 'foo' * 100
            self.bar = 'bar' * 100

    class DummyClassTwo(DummyClassOne):
        def __init__(self):
            super().__init__()
            self.baz = 'baz' * 100
            self.blah = 'blah' * 100

    class DummyClassThree(DummyClassTwo):
        def __init__(self):
            super().__init__()
            self.a_number = 5
            self.when = 'when'
            self._in = 'in'
            self.the = 'the'
            self.course = 'course'
            self.of = 'of'
            self.human = 'human'
            self.events = 'events'

    class DummyClassFour(DummyClassThree):
        pass

    TEST_FILTERS = {
        str: ALWAYS,
        DummyClassOne: {
            'foo': ALWAYS,
            'bar': ALWAYS
        },
        DummyClassTwo: {
            'baz': lambda inst: inst.baz.upper(),
            'blah': lambda inst: inst.blah.upper()
        },
        DummyClassThree: {
            'foo': NEVER,
            'a_number': lambda inst: inst.a_number + 5,
            'when': ALWAYS,
            '_in': ALWAYS,
            'the': ALWAYS,
            'course': NEVER,
            'of': ALWAYS,
            'human': ALWAYS,
            'events': ALWAYS
        }
    }

    def test_filter_large_collection(self):
        data = [EndpointFilteringTestCase.DummyClassFour() for _ in range(1000)]
        filtered_data = filtering.apply_filters_to_object(data, EndpointFilteringTestCase.TEST_FILTERS)
        self.assertEqual(1000, len(filtered_data))
        self.assertFalse('course' in filtered_data[0])
        self.assertTrue('when' in filtered_data[0])
        self.assertEqual(10, filtered_data[0]['a_number'])
        self.assertEqual('BLAH' * 100, filtered_data[0]['blah'])

    def test_init_without_consumer_attributes(self):
        class _TestEndpoint(machinery.BaseEndpointDefinition):
            @property
            def resource(self):
                return {}

        endpoint = _TestEndpoint()
        endpoint_binder = machinery.EndpointBinder(endpoint)
        self.assertEqual(endpoint_binder.consumer_attributes, [])

    def test_create_bound_endpoint_with_url_and_adhoc_query_fields(self):
        req = django.test.RequestFactory().get('/')
        req.consumer = self.consumer
        testutils.OAuthClientHandler._build_request(req)
        params = req.GET.copy()
        params['adhoc_field__lt'] = 'bar'
        req.GET = params

        class _TestEndpoint(machinery.EndpointDefinition):
            url_field = machinery.url_field()
            adhoc_field = machinery.adhoc_queryset()

            def is_authorized(self):
                return True

            @property
            def resource(self):
                return {
                    'adhoc_field': self.adhoc_field,
                    'url_field': self.url_field,
                }

        bound_endpoint = _bind_endpoint(_TestEndpoint, req, url_field='baz')

        status, data = bound_endpoint.get_response()
        self.assertEqual(status, http.HTTPStatus.OK)
        self.assertEqual(data['adhoc_field'], {'adhoc_field__lt': 'bar'})
        self.assertEqual(data['url_field'], 'baz')

    @mock.patch('django_declarative_apis.machinery.EndpointBinder._validate_endpoint')
    def test_create_bound_endpoint_exception_raised(self, mock_validate_endpoint):
        mock_validate_endpoint.side_effect = Exception('something bad happened')
        req = django.test.RequestFactory().get('/')
        req.consumer = self.consumer
        testutils.OAuthClientHandler._build_request(req)

        class _TestEndpoint(machinery.EndpointDefinition):
            @property
            def resource(self):
                return {}

        bound_endpoint = _bind_endpoint(_TestEndpoint, req)

        try:
            status, data = bound_endpoint.get_response()
            self.fail('Should have failed')
        except Exception as err:
            _, err_inst, _ = bound_endpoint.validation_exc_info
            self.assertEqual(err_inst, err)

    def test_validate_endpoint_unauthorized(self):
        req = django.test.RequestFactory().get('/')
        req.consumer = self.consumer
        testutils.OAuthClientHandler._build_request(req)

        class _TestEndpoint(machinery.EndpointDefinition):
            required_field = machinery.url_field(required=True)

            @property
            def resource(self):
                return {}

        bound_endpoint = _bind_endpoint(_TestEndpoint, req)

        try:
            bound_endpoint.get_response()
            self.fail('this should have failed')
        except errors.ClientErrorForbidden as err:
            expected_code, expected_msg = errors.FORBIDDEN
            self.assertEqual(err.save_changes, False)
            self.assertEqual(err.error_code, expected_code)
            self.assertEqual(err.error_message, expected_msg)
            self.assertEqual(err.status_code, http.HTTPStatus.FORBIDDEN)

    def test_validate_endpoint_missing_consumer_error(self):
        req = self.create_request()

        class _TestEndpoint(machinery.EndpointDefinition):
            consumer = machinery.consumer_attribute()

            @property
            def resource(self):
                return {}

        bound_endpoint = _bind_endpoint(_TestEndpoint, req)

        try:
            bound_endpoint.get_response()
            self.fail('this should have failed')
        except errors.ClientErrorForbidden as err:
            expected_code, expected_msg = errors.FORBIDDEN
            self.assertEqual(err.save_changes, False)
            self.assertEqual(err.error_code, expected_code)
            self.assertEqual(err.error_message, expected_msg)
            self.assertEqual(err.status_code, http.HTTPStatus.FORBIDDEN)

    def test_get_response_missing_required_attributes_error(self):
        req = self.create_request()

        class _TestEndpoint(machinery.EndpointDefinition):
            required_a = machinery.field(required=True)
            required_b = machinery.field(required=True)

            @property
            def resource(self):
                return {}

        bound_endpoint = _bind_endpoint(_TestEndpoint, req)
        try:
            bound_endpoint.get_response()
        except errors.ClientErrorMissingFields as err:
            expected_code, expected_msg = errors.MISSING_FIELDS
            self.assertFalse(err.save_changes)
            self.assertTrue(err.error_message.startswith(expected_msg))
            self.assertTrue(all(name in err.error_message for name in ('required_a', 'required_b')))
            self.assertEqual(err.error_code, expected_code)
            self.assertEqual(err.status_code, http.HTTPStatus.BAD_REQUEST)

    def test_get_response_invalid_field_type_error(self):
        req = self.create_request(method='POST', body={'int_field_a': 'foo', 'int_field_b': 'bar'})

        class _TestEndpoint(machinery.EndpointDefinition):
            int_field_a = machinery.field(type=int, required=True)
            int_field_b = machinery.field(type=int, required=True)

            def is_authorized(self):
                return True

            @property
            def resource(self):
                return {}

        bound_endpoint = _bind_endpoint(_TestEndpoint, req)

        try:
            bound_endpoint.get_response()
            self.fail('should have failed here')
        except errors.ClientErrorInvalidFieldValues as err:
            expected_code, expected_msg = errors.INVALID_FIELD_VALUES
            self.assertFalse(err.save_changes)
            self.assertTrue(err.error_message.startswith(expected_msg))
            self.assertTrue('int_field_a' in err.error_message)
            self.assertTrue('int_field_b' in err.error_message)
            self.assertEqual(err.error_code, expected_code)
            self.assertEqual(err.status_code, http.HTTPStatus.BAD_REQUEST)

    def test_get_response_not_found_error(self):
        req = self.create_request()

        class _TestEndpoint(machinery.EndpointDefinition):
            def is_authorized(self):
                raise django.core.exceptions.ObjectDoesNotExist

            @property
            def resource(self):
                return {}

        bound_endpoint = _bind_endpoint(_TestEndpoint, req)

        try:
            bound_endpoint.get_response()
            self.fail('should have failed')
        except errors.ClientErrorNotFound as err:
            self.assertFalse(err.save_changes)
            self.assertEqual(err.error_message, http.HTTPStatus.NOT_FOUND.phrase)
            self.assertEqual(err.error_code, http.HTTPStatus.NOT_FOUND)
            self.assertEqual(err.status_code, http.HTTPStatus.NOT_FOUND)

    def test_get_response_throttled_error(self):
        req = self.create_request()

        class _TestEndpoint(machinery.EndpointDefinition):
            def is_authorized(self):
                return True

            def rate_limit_key(self):
                return 'foo'

            @property
            def resource(self):
                return {}

        with mock.patch('django_declarative_apis.machinery.rate_limit_exceeded') as patch_throttle:
            patch_throttle.return_value = True
            bound_endpoint = _bind_endpoint(_TestEndpoint, req)

        try:
            bound_endpoint.get_response()
        except errors.ClientErrorRequestThrottled as err:
            expected_code, expected_message = errors.REQUEST_THROTTLED
            self.assertFalse(err.save_changes)
            self.assertEqual(err.error_code, expected_code)
            self.assertEqual(err.error_message, expected_message)
            self.assertEqual(err.status_code, http.HTTPStatus.TOO_MANY_REQUESTS)


class EndpointRequestLifecycleManagerTestCase(testutils.RequestCreatorMixin, unittest.TestCase):
    def test_process_request_and_get_response_success(self):
        req = self.create_request()
        expected_data = {'foo': 'bar'}

        class _TestEndpoint(machinery.EndpointDefinition):
            def is_authorized(self):
                return True

            @property
            def resource(self):
                return expected_data

        manager = machinery._EndpointRequestLifecycleManager(_TestEndpoint)
        status, data = manager.process_request_and_get_response(req)
        self.assertEqual(status, http.HTTPStatus.OK)
        self.assertEqual(data, expected_data)


class EndpointDefinitionTestCase(testutils.RequestCreatorMixin, unittest.TestCase):
    def test_is_permitted(self):
        self.consumer.type = dda_models.BaseConsumer.TYPE_READ_ONLY
        self.consumer.save()

        class _TestEndpoint(machinery.EndpointDefinition):
            def is_authorized(self):
                return True

            @property
            def resource(self):
                return {}

        # client should be able to access read endpoints
        req = self.create_request()
        bound_endpoint = _bind_endpoint(_TestEndpoint, req)
        status, _ = bound_endpoint.get_response()
        self.assertEqual(status, http.HTTPStatus.OK)

        # but not write
        req = self.create_request(method='POST')
        bound_endpoint = _bind_endpoint(_TestEndpoint, req)
        self.assertRaises(errors.ClientErrorForbidden, bound_endpoint.get_response)

    def test_is_permitted_readonly(self):
        self.consumer.type = dda_models.BaseConsumer.TYPE_READ_ONLY
        self.consumer.save()

        class _TestReadOnlyEndpoint(machinery.EndpointDefinition):
            is_read_only = True

            def is_authorized(self):
                return True

            @property
            def resource(self):
                return {}

        req = self.create_request(method='POST')
        bound_endpoint = _bind_endpoint(_TestReadOnlyEndpoint, req)
        status, _ = bound_endpoint.get_response()
        self.assertEqual(status, http.HTTPStatus.OK)

    def test_misconfigured_consumer_cant_access_resource(self):
        self.consumer.type = 'invalid'
        self.consumer.save()

        class _TestEndpoint(machinery.EndpointDefinition):
            def is_authorized(self):
                return True

            @property
            def resource(self):
                return {}

        req = self.create_request()
        bound_endpoint = _bind_endpoint(_TestEndpoint, req)
        self.assertRaises(errors.ClientErrorForbidden, bound_endpoint.get_response)

        req = self.create_request(method='POST')
        bound_endpoint = _bind_endpoint(_TestEndpoint, req)
        self.assertRaises(errors.ClientErrorForbidden, bound_endpoint.get_response)


class ResourceUpdateEndpointDefinitionTestCase(testutils.RequestCreatorMixin, django.test.TestCase):
    def test_mutate_resource(self):
        obj = tests.models.TestModel(int_field=24)
        obj.save()

        class _TestEndpoint(machinery.ResourceUpdateEndpointDefinition):
            resource_model = tests.models.TestModel
            int_field = machinery.resource_field(type=int)

            def __init__(self, *args, **kwargs):
                self.resource_id = obj.id
                super().__init__(*args, **kwargs)

            def is_authorized(self):
                return True

            @machinery.endpoint_resource(
                type=tests.models.TestModel,
                filter={tests.models.TestModel: {'int_field': filtering.ALWAYS}}
            )
            def resource(self):
                return super().resource

            @property
            def response(self):
                self.resource.save()
                return self.resource

        req = self.create_request(method='POST', body={'int_field': 42}, use_auth_header_signature=True)
        bound_endpoint = _bind_endpoint(_TestEndpoint, req)
        status, resp = bound_endpoint.get_response()
        self.assertEqual(status, http.HTTPStatus.OK)
        self.assertEqual(resp, {'int_field': 42})


class ResourceCreationMixinTestCase(unittest.TestCase):
    def test_status(self):
        class Test(machinery.ResourceCreationMixin):
            pass

        self.assertEqual(Test().http_status, http.HTTPStatus.CREATED)


class MySpecialException(Exception):
    pass


class _TestEndpoint(machinery.EndpointDefinition):
    def __init__(self, expected_response, *args, **kwargs):
        super(_TestEndpoint, self).__init__(*args, **kwargs)
        self.expected_response = expected_response

    @machinery.endpoint_resource(type=tests.models.TestModel, filter={'foo': 'bar'})
    def resource(self):
        return tests.models.TestModel(int_field=0)

    @property
    def response(self):
        return self.expected_response

    @machinery.deferrable_task(always_defer=True)
    @staticmethod
    def deferred_task(inst):
        assert inst is not None
        _TestEndpoint.semaphore['status'] = 'deferred task executed'

    @machinery.deferrable_task(always_defer=True, retries=2)
    @staticmethod
    def deferred_task_with_retry(inst):
        assert inst is not None
        retry_key = 'retry_count'
        _TestEndpoint.semaphore[retry_key] += 1
        if _TestEndpoint.semaphore[retry_key] < 3:
            raise Exception("failing so task will retry")

    @machinery.deferrable_task(always_defer=True, retries=2, retry_exception_filter=(MySpecialException,))
    @staticmethod
    def deferred_task_with_retry_and_filtered_exceptions(inst):
        assert inst is not None
        retry_key = 'filtered_retry_count_1'
        _TestEndpoint.semaphore[retry_key] += 1
        if _TestEndpoint.semaphore[retry_key] < 3:
            raise MySpecialException("failing so task will retry")

    @machinery.deferrable_task(always_defer=True, retries=2, retry_exception_filter=(MySpecialException,))
    @staticmethod
    def deferred_task_with_retry_and_filtered_exceptions_fail(inst):
        # this task specifies an exception filter list, but throws the wrong exception, so should not retry
        assert inst is not None
        retry_key = 'filtered_retry_count_2'
        _TestEndpoint.semaphore[retry_key] += 1
        if _TestEndpoint.semaphore[retry_key] < 3:
            raise Exception("failing so task will retry")


class DeferrableTaskTestCase(django.test.TestCase):
    def setUp(self):
        _TestEndpoint.semaphore = {
            'status': None,
            'retry_count': 0,
            'filtered_retry_count_1': 0,
            'filtered_retry_count_2': 0
            }

    def test_get_response_kombu_error_retried(self):
        expected_response = {'foo': 'bar'}
        endpoint = _TestEndpoint(expected_response)
        manager = machinery.EndpointBinder.BoundEndpointManager(
            machinery._EndpointRequestLifecycleManager(endpoint),
            endpoint
        )

        conf = tasks.future_task_runner.app.conf
        old_val = conf['task_always_eager']
        conf['task_always_eager'] = True

        cache.set(tasks.JOB_COUNT_CACHE_KEY, 0)

        with mock.patch('django_declarative_apis.machinery.tasks.future_task_runner.apply_async') as mock_apply:
            exceptions = iter([kombu.exceptions.OperationalError, kombu.exceptions.OperationalError])
            def _side_effect(*args, **kwargs):
                try:
                    raise next(exceptions)
                except StopIteration:
                    return future_task_runner.apply(*args, **kwargs)
            mock_apply.side_effect = _side_effect

            try:
                resp = manager.get_response()
            finally:
                conf['task_always_eager'] = old_val

        self.assertEqual(resp, (http.HTTPStatus.OK, expected_response))
        self.assertTrue(cache.get(tasks.JOB_COUNT_CACHE_KEY) != 0)

        self.assertEqual('deferred task executed', _TestEndpoint.semaphore['status'])

    def test_async_task_falls_back_to_synchronous_when_configured(self):
        expected_response = {'foo': 'bar'}

        endpoint = _TestEndpoint(expected_response)
        manager = machinery.EndpointBinder.BoundEndpointManager(
            machinery._EndpointRequestLifecycleManager(endpoint),
            endpoint
        )

        conf = tasks.future_task_runner.app.conf
        old_val = conf['task_always_eager']
        conf['task_always_eager'] = True

        with mock.patch('django_declarative_apis.machinery.tasks.future_task_runner.apply_async') as mock_apply_async:
            mock_apply_async.side_effect = kombu.exceptions.OperationalError

            cache.set(tasks.JOB_COUNT_CACHE_KEY, 0)

            with self.settings(DECLARATIVE_ENDPOINT_TASKS_SYNCHRONOUS_FALLBACK=True):
                try:
                    resp = manager.get_response()
                except kombu.exceptions.OperationalError:
                    self.fail('OperationalError should not have been triggered')
                finally:
                    conf['task_always_eager'] = old_val

        self.assertEqual('deferred task executed', _TestEndpoint.semaphore['status'])

    def test_force_synchronous_tasks(self):
        expected_response = {'foo': 'bar'}
        endpoint = _TestEndpoint(expected_response)
        manager = machinery.EndpointBinder.BoundEndpointManager(
            machinery._EndpointRequestLifecycleManager(endpoint),
            endpoint
        )

        conf = tasks.future_task_runner.app.conf
        old_val = conf['task_always_eager']
        conf['task_always_eager'] = True

        cache.set(tasks.JOB_COUNT_CACHE_KEY, 0)

        with mock.patch('django_declarative_apis.machinery.tasks.future_task_runner.apply_async') as mock_apply:
            mock_apply.side_effect = kombu.exceptions.OperationalError

            with self.settings(DECLARATIVE_ENDPOINT_TASKS_FORCE_SYNCHRONOUS=True):
                try:
                    resp = manager.get_response()
                except kombu.exceptions.OperationalError:
                    self.fail('OperationalError should not have been triggered')
                finally:
                    conf['task_always_eager'] = old_val

        self.assertEqual(0, mock_apply.call_count)
        self.assertEqual('deferred task executed', _TestEndpoint.semaphore['status'])

    def test_get_response_kombu_error_attempts_exceeded(self):
        expected_response = {'foo': 'bar'}
        endpoint = _TestEndpoint(expected_response)
        manager = machinery.EndpointBinder.BoundEndpointManager(
            machinery._EndpointRequestLifecycleManager(endpoint),
            endpoint
        )

        conf = tasks.future_task_runner.app.conf
        old_val = conf['task_always_eager']
        conf['task_always_eager'] = True

        cache.set(tasks.JOB_COUNT_CACHE_KEY, 0)

        with mock.patch('django_declarative_apis.machinery.tasks.future_task_runner.apply_async') as mock_apply:
            exceptions = iter([
                kombu.exceptions.OperationalError,
                kombu.exceptions.OperationalError,
                kombu.exceptions.OperationalError,
            ])
            def _side_effect(*args, **kwargs):
                try:
                    raise next(exceptions)
                except StopIteration:
                    return future_task_runner.apply(*args, **kwargs)
            mock_apply.side_effect = _side_effect

            try:
                resp = manager.get_response()
                self.fail('should have triggered a kombu.exceptions.OperationalError')
            except kombu.exceptions.OperationalError:
                pass
            finally:
                conf['task_always_eager'] = old_val

        self.assertIsNone(_TestEndpoint.semaphore['status'])

    def test_get_response_success(self):
        expected_response = {'foo': 'bar'}

        endpoint = _TestEndpoint(expected_response)
        manager = machinery.EndpointBinder.BoundEndpointManager(
            machinery._EndpointRequestLifecycleManager(endpoint),
            endpoint
        )

        # can't use mock.patch.dict here because it doesn't implement the api that the unpatcher expects
        conf = tasks.future_task_runner.app.conf
        old_val = conf['task_always_eager']
        conf['task_always_eager'] = True

        cache.set(tasks.JOB_COUNT_CACHE_KEY, 0)

        try:
            resp = manager.get_response()
        finally:
            conf['task_always_eager'] = old_val

        self.assertEqual(resp, (http.HTTPStatus.OK, expected_response))
        self.assertTrue(cache.get(tasks.JOB_COUNT_CACHE_KEY) != 0)

        self.assertEqual('deferred task executed', _TestEndpoint.semaphore['status'])
        self.assertEqual(3, _TestEndpoint.semaphore['retry_count'])
        self.assertEqual(3, _TestEndpoint.semaphore['filtered_retry_count_1'])
        self.assertEqual(1, _TestEndpoint.semaphore['filtered_retry_count_2'])


        _TestEndpoint.semaphore = {
            'status': None,
            'retry_count': 0,
            'filtered_retry_count_1': 0,
            'filtered_retry_count_2': 0
            }
        
    def test_deferrable_methods_must_be_static(self):
        try:
            machinery.deferrable_task()(DeferrableTaskTestCase.test_deferrable_methods_must_be_static)
            self.fail('Should have raised assertionerror')
        except AssertionError as err:
            self.assertTrue('Deferrable task methods MUST be staticmethods' in str(err))
        except Exception:
            self.fail('should have raised assertionerror')


class ResourceAsyncJobTestCase(django.test.TestCase):
    def test_resource_async_job(self):
        resource = tests.models.TestModel()
        resource.int_field = 1
        resource.save()

        conf = tasks.future_task_runner.app.conf
        old_val = conf['task_always_eager']
        conf['task_always_eager'] = True

        tasks.schedule_resource_task_runner(resource.mutate_action)

        conf['task_always_eager'] = old_val

        resource.refresh_from_db()
        self.assertEqual(2, resource.int_field)


def _bind_endpoint(endpoint_cls, request, url_field=None):
    endpoint = endpoint_cls()
    binder = machinery.EndpointBinder(endpoint_cls)
    return binder.create_bound_endpoint(
        machinery._EndpointRequestLifecycleManager(endpoint), request, url_field=url_field
    )



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2018-05-23 10:03
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_declarative_apis', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='oauthconsumer',
            name='name',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='oauthconsumer',
            name='type',
            field=models.CharField(choices=[('RO', 'Read Only'), ('RW', 'Read/Write')], db_index=True, default='RW', max_length=2),
        ),
    ]



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

from importlib import import_module

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .machinery import BehavioralEndpointDefinitionRouter

import django_declarative_apis.authentication

# don't require resources unless this adapter is in use
try:
    from django_declarative_apis.resources.resource import Resource
except ImportError as e:  # pragma: nocover
    import traceback
    traceback.print_exc()
    Resource = object


class BaseHandler(object):
    defined_methods = {'get', 'put', 'patch', 'post', 'delete'}


class EndpointHandler(object):
    """
    Glue for combining the new-style endpoint definitions into the old-style piston handler

    """

    def __init__(self, **kwargs):
        super(EndpointHandler, self).__init__()

        self.method_handlers = {}
        for method, handler in kwargs.items():
            if method not in BaseHandler.defined_methods:
                raise TypeError('Unexpected keyword argument {0}: valid arguments are {1}'
                                .format(method, BaseHandler.defined_methods))

            if isinstance(handler, (list, tuple)):
                self.method_handlers[method.upper()] = BehavioralEndpointDefinitionRouter(*handler)
            else:
                self.method_handlers[method.upper()] = BehavioralEndpointDefinitionRouter(handler)

        self.allowed_methods = self.method_handlers.keys()

    def __call__(self, *args, **kwargs):
        return self

    def handle_request(self, method, *args, **kwargs):
        return self.method_handlers[method](*args, **kwargs)

    @property
    def documentation(self):
        return {method: handler.documentation for method, handler in self.method_handlers.items()}


class EndpointResource(Resource):
    def __init__(self, authentication=None, **kwargs):
        super(EndpointResource, self).__init__(EndpointHandler(**kwargs))

        if authentication is not None:
            django_declarative_apis.authentication.validate_authentication_config(authentication)
            self.authentication = authentication


def resource_adapter(*args, **kwargs):
    setting_name = 'DECLARATIVE_ENDPOINT_RESOURCE_ADAPTER'
    adapter_name = getattr(settings, setting_name, None)
    if not adapter_name:
        raise ImproperlyConfigured(setting_name)

    name_components = adapter_name.split('.')
    module_name = '.'.join(name_components[:-1])
    module = import_module(module_name)
    class_name = name_components[-1]
    adapter_class = getattr(module, class_name)
    return adapter_class(*args, **kwargs)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-29 15:12
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='OauthConsumer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.CharField(max_length=100, null=True)),
                ('key', models.CharField(db_index=True, max_length=18)),
                ('secret', models.CharField(max_length=32)),
                ('rsa_public_key_pem', models.TextField(blank=True, null=True)),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import logging

from django.conf import settings
from django.core.cache import cache
from oauthlib.oauth1 import RequestValidator

from django_declarative_apis import models as oauth_models


class DjangoRequestValidator(RequestValidator):

    TIMESTAMP_THRESHOLD = 300

    def __init__(self, request, *args, **kwargs):
        super(DjangoRequestValidator, self).__init__(*args, **kwargs)
        self.request = request
        self.consumer = None
        self.validation_error_message = ''

    @property
    def enforce_ssl(self):
        return getattr(settings, 'REQUIRE_HTTPS_FOR_OAUTH', True)

    def check_client_key(self, client_key):
        return bool(client_key)

    def check_nonce(self, nonce):
        return bool(nonce)  # we didn't enforce any minimum length on nonces previously...

    def validate_timestamp_and_nonce(self, client_key, timestamp, nonce,
                                     request, request_token=None, access_token=None):
        request_token = request_token or ''
        access_token = access_token or ''
        cache_key = '{0}:{1}:{2}:{3}'.format(client_key, request_token, access_token, nonce)
        cache_created = cache.add(cache_key, True, self.TIMESTAMP_THRESHOLD * 2)
        if not cache_created:
            self.validation_error_message = 'nonce_used'
        return cache_created

    def validate_client_key(self, client_key, request):
        self.consumer = oauth_models.get_consumer(client_key)
        if self.consumer is None:
            self.validation_error_message = 'consumer_key_unknown'
            logging.error('invalid consumer')
            return False
        return True

    def get_client_secret(self, client_key, request):
        try:
            if self.consumer.rsa_public_key_pem:
                return None
            else:
                return self.consumer.secret
        except Exception as e:
            logging.error("This should never happen, since consumer is already validated")
            return ''

    def get_rsa_key(self, client_key, request):
        try:
            return self.consumer.rsa_public_key_pem
        except Exception as e:
            logging.error("This should never happen, since consumer is already validated")
            return ''

    @property
    def dummy_client(self):
        class DummyClient(object):
            def __init__(self, *args, **kwargs):
                self.secret = ''
                self.key = ''
                self.rsa_public_key_base64 = ''
        return DummyClient()





#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import itertools
import sys
import http.client

import django
from django.conf import settings
from django.db import models
from django.http import HttpResponse

from django_declarative_apis.machinery.filtering import apply_filters_to_object
from django_declarative_apis.resources.utils import HttpStatusCode
from django_declarative_apis.models import BaseConsumer
from . import errors
from .attributes import *
from .utils import rate_limit_exceeded, locate_object


# TODO:
# * Make it generically handle database write failures (updating the http_status to be 5XX)
# * Create new error for deadline exceeded and catch it in the same place as writes
# * Make deferred tasks actually run deferred


class EndpointResourceAttribute(EndpointAttribute):
    def __init__(self, type, filter=None, returns_list=False, **kwargs):
        super(EndpointResourceAttribute, self).__init__(**kwargs)
        self.type = type
        self.filter = filter
        self.func = None
        self.returns_list = returns_list

    def __call__(self, func):
        self.func = func
        return self

    def get_instance_value(self, owner_instance, owner_class):
        if not owner_instance:
            return self
        try:
            value = self.func(owner_instance)
        except django.core.exceptions.ObjectDoesNotExist:
            raise errors.ClientErrorNotFound("{0} instance not found".format(self.type.__name__))

        if value.__class__ == dict:
            return value

        if not getattr(value, '_api_filter', False):
            value._api_filter = self.filter

        return value


class EndpointResponseAttribute(EndpointAttribute):
    def __init__(self, type, filter=None, **kwargs):
        super(EndpointResponseAttribute, self).__init__(**kwargs)
        self.type = type
        self.filter = filter
        self.func = None

    def __call__(self, func):
        self.func = func
        return self

    def get_instance_value(self, owner_instance, owner_class):
        if not owner_instance:
            return self
        value = self.func(owner_instance)
        if not getattr(value, '_api_filter', False):
            if self.filter:
                value._api_filter = self.filter
        return value


class EndpointDefinitionMeta(abc.ABCMeta, metaclass=abc.ABCMeta):
    def __init__(cls, class_name, bases=None, dict=None):
        super(EndpointDefinitionMeta, cls).__init__(class_name, bases, dict)

        # This metaclass sets EndpointAttributeDiscriptor's names if they haven't otherwise been set
        # This will walk parent classes as well so that attributes can be defined through inheritance
        ancestor_attribs = (ancestor.__dict__.items() for ancestor in cls.mro())
        for name, attribute in itertools.chain(dict.items(), *ancestor_attribs):
            try:
                if not attribute.name:
                    attribute.name = name
            except AttributeError as e:  # noqa
                pass


class EndpointBinder(object):
    class BoundEndpointManager(object):
        def __init__(self, manager, bound_endpoint):
            self.manager = manager
            self.bound_endpoint = bound_endpoint
            self.binding_exc_info = None
            self.validation_exc_info = None

        def get_response(self):
            error = self.binding_exc_info or self.validation_exc_info
            if error:
                exc_type, exc_value, exc_traceback = error
                if isinstance(exc_value, errors.ClientError):
                    logging.warning(exc_value.error_message)
                else:
                    logging.error(str(exc_value.args) + '\n' + str(exc_traceback))

                raise exc_value.with_traceback(exc_traceback)

            resource = self.bound_endpoint.resource

            if hasattr(resource, 'is_dirty'):
                if resource and resource.is_dirty(check_relationship=True):
                    resource.save()

            endpoint_tasks = sorted(self.manager.endpoint_tasks, key=lambda t: t.priority)
            immediate_tasks = filter(lambda t: not isinstance(t, DeferrableEndpointTask), endpoint_tasks)
            deferred_tasks = filter(lambda t: isinstance(t, DeferrableEndpointTask), endpoint_tasks)
            try:
                for immediate_task in immediate_tasks:
                    immediate_task.run(self.bound_endpoint)

            except errors.ClientError as ce:
                if ce.save_changes and resource and resource.is_dirty():
                    resource.save()
                raise

            if hasattr(resource, 'is_dirty'):
                if resource and resource.is_dirty(check_relationship=True):
                    resource.save()

            for deferred_task in deferred_tasks:
                deferred_task.run(self.bound_endpoint)

            if getattr(resource, '_api_filter', False):
                filter_def = resource._api_filter
            else:
                filter_def = self.bound_endpoint.response_filter

            data = self.bound_endpoint.response
            status_code = self.bound_endpoint.http_status

            if isinstance(data, HttpResponse):
                if 200 <= status_code <= 299:
                    return status_code, data
                else:
                    raise HttpStatusCode(data)
            elif isinstance(data, dict):
                result = {}
                for key, value in data.items():
                    if isinstance(value, (list, tuple, models.query.QuerySet)):
                        result[key] = []
                        for item in value:
                            result[key].append(apply_filters_to_object(item, filter_def))
                    else:
                        result[key] = value

                return status_code, result
            else:
                return status_code, apply_filters_to_object(data,
                                                            filter_def,
                                                            self.bound_endpoint.request.META.get('HTTP_X_EXPAND'))

    def __init__(self, endpoint_definition):
        super(EndpointBinder, self).__init__()
        self.endpoint_definition = endpoint_definition
        self.endpoint_attributes = endpoint_definition.get_endpoint_attributes()
        self.request_properties = endpoint_definition.get_request_properties()
        self.required_request_properties = endpoint_definition.get_required_request_properties()
        try:
            self.consumer_attributes = endpoint_definition.get_consumer_attributes()
        except AttributeError:
            self.consumer_attributes = []
        self.request_fields = endpoint_definition.get_request_fields()
        self.required_request_fields = endpoint_definition.get_required_request_fields()
        self.endpoint_tasks = endpoint_definition.get_tasks()
        self.url_fields = endpoint_definition.get_url_fields()
        self.adhoc_queries = endpoint_definition.get_adhoc_queries()

    def create_bound_endpoint(self, manager, request, *args, **kwargs):
        endpoint = self.endpoint_definition()

        for url_field in self.url_fields:
            if (url_field.api_name or url_field.name) in kwargs:
                url_field.set_value(kwargs.get(url_field.api_name or url_field.name))

        for adhoc_query_field in self.adhoc_queries:
            adhoc_query_field.set_value({
                key: val for (key, val) in request.GET.items()
                if key.startswith(adhoc_query_field.name)
            })

        # Bind the request object within the instance (this allows RequestProperties to access the request
        # without the endpoint definition having direct access to it)
        RequestProperty.bind_request_to_instance(endpoint, request)

        bound_endpoint_manager = EndpointBinder.BoundEndpointManager(manager, endpoint)

        try:
            self._bind_endpoint(endpoint)
        except Exception as e:
            bound_endpoint_manager.binding_exc_info = sys.exc_info()
            return bound_endpoint_manager

        try:
            self._validate_endpoint(endpoint)
        except Exception as e:
            bound_endpoint_manager.validation_exc_info = sys.exc_info()

        return bound_endpoint_manager

    def _bind_endpoint(self, endpoint):
        # Access all request properties (this validates a request using the definition and caches the values)
        extra_error_message = ''
        missing_required_properties = []
        invalid_value_properties = []
        for request_property in self.request_properties:
            try:
                value = getattr(endpoint, request_property.name)
                if value is None and request_property.required:
                    if isinstance(request_property, ConsumerAttribute):
                        # A missing required consumer attribute should fail quickly as forbidden
                        raise errors.ClientErrorForbidden()
                    else:
                        # Otherwise collect missing properties and report them all together
                        missing_required_properties.append(request_property)
            except errors.ClientErrorMissingFields as mfe:  # TODO: seems unreachable
                extra_error_message += mfe.error_message  # pragma: nocover
            except (ValueError, errors.ClientErrorInvalidFieldValues) as ve:
                # Collect invalid values and report them all together
                invalid_value_properties.append(request_property)  # pragma: nocover

        if missing_required_properties or extra_error_message:
            raise errors.ClientErrorMissingFields([property.name for property in missing_required_properties],
                                                  extra_message=extra_error_message)

        if invalid_value_properties:
            raise errors.ClientErrorInvalidFieldValues(
                    [request_property.name for request_property in invalid_value_properties])

    def _validate_endpoint(self, endpoint):
        # Run standard validators
        try:
            if not (endpoint.is_authorized() and endpoint.is_permitted() and endpoint.is_valid()):
                raise errors.ClientErrorForbidden(additional_info=getattr(endpoint, '_validation_error_message', None))

        except django.core.exceptions.ObjectDoesNotExist:
            raise errors.ClientErrorNotFound()

        # check ratelimit
        rate_limit_key = endpoint.rate_limit_key()
        if (rate_limit_key is not None) and rate_limit_exceeded(rate_limit_key, endpoint.rate_limit_period()):
            raise errors.ClientErrorRequestThrottled()


class _EndpointRequestLifecycleManager(object):
    def __init__(self, endpoint_definition):
        super(_EndpointRequestLifecycleManager, self).__init__()
        self.endpoint_definition = endpoint_definition
        self.binder = EndpointBinder(endpoint_definition)
        self.endpoint_tasks = endpoint_definition.get_tasks()

    def bind_endpoint_to_request(self, request, *args, **kwargs):
        return self.binder.create_bound_endpoint(self, request, *args, **kwargs)

    def process_request_and_get_response(self, request, *args, **kwargs):
        bound_endpoint = self.bind_endpoint_to_request(request, *args, **kwargs)
        return bound_endpoint.get_response()

    def __str__(self):  # pragma: nocover
        return self.endpoint_definition.__name__


class BehavioralEndpointDefinitionRouter(object):
    def __init__(self, *endpoint_definitions):
        super(BehavioralEndpointDefinitionRouter, self).__init__()
        self.endpoint_definitions = endpoint_definitions
        self.endpoint_managers = [
            _EndpointRequestLifecycleManager(endpoint) for endpoint in endpoint_definitions
            ]
        self.endpoint_manager_names = '({0})'.format(','.join(map(lambda e: e.__name__, endpoint_definitions)))

    def bind_endpoint_to_request(self, request, *args, **kwargs):
        bound_endpoint = None
        for candidate_endpoint_manager in self.endpoint_managers:
            bound_endpoint = candidate_endpoint_manager.bind_endpoint_to_request(request, *args, **kwargs)
            if bound_endpoint.binding_exc_info is None:
                break
        return bound_endpoint

    def process_request_and_get_response(self, request, *args, **kwargs):
        try:
            bound_endpoint = self.bind_endpoint_to_request(request, *args, **kwargs)
            logging.info("Processing request with handler %s", bound_endpoint.bound_endpoint.__class__.__name__)
            return bound_endpoint.get_response()

        except errors.ApiError:
            raise
        except Exception as e:  # pragma: nocover
            raise errors.ServerError() from e

    def __call__(self, *args, **kwargs):
        return self.process_request_and_get_response(*args, **kwargs)

    def __str__(self):  # pragma: nocover
        return self.endpoint_manager_names

    @property
    def documentation(self):
        return [x.documentation() for x in self.endpoint_definitions]


class EndpointDefinitionMixin(metaclass=EndpointDefinitionMeta):
    pass


class BaseEndpointDefinition(metaclass=EndpointDefinitionMeta):
    @abc.abstractmethod
    def is_authorized(self):
        """ Authorization check. Should be overridden by endpoint definition implementations.

        Returns:
            ``bool``: Whether or not the user should be able to access the resource. Defaults to ``False``.
        """
        return False

    def is_permitted(self):
        return True

    def is_valid(self):
        return True

    def rate_limit_key(self):
        """
        Should return a unique key that is used for rate-limiting requests to this endpoint.
        Return None if the request should not be rate-limited
        """
        return None

    def rate_limit_period(self):
        """
        number of seconds to enforce between requests with the same rate_limit_key
        """
        return 1

    @property
    def response_filter(self):
        filter_def_name = getattr(settings, 'DECLARATIVE_ENDPOINT_DEFAULT_FILTERS', None)
        if filter_def_name:
            filter_def = locate_object(filter_def_name)
        else:
            filter_def = {}
        return filter_def

    @property
    def http_status(self):
        return http.client.OK

    @property
    @abc.abstractmethod
    def resource(self):
        """ The instance of a resource. Should either be a ``dict`` or instance of a Django Model or QuerySet.

        This property *must* be implemented by all endpoint definitions.
        """
        raise NotImplementedError('Endpoints must implement self.resource property')

    @property
    def response(self):
        return self.resource

    @classmethod
    def get_endpoint_attributes(cls):
        endpoint_attributes = filter(lambda attribute: isinstance(attribute, EndpointAttribute),
                                     [getattr(cls, name) for name in dir(cls)])
        return sorted(endpoint_attributes, key=lambda attribute: attribute.attribute_number)

    @classmethod
    def get_request_properties(cls):
        return list(filter(lambda attribute: isinstance(attribute, RequestProperty),
                           cls.get_endpoint_attributes()))

    @classmethod
    def get_required_request_properties(cls):
        return list(filter(lambda property: property.required,
                           cls.get_request_properties()))

    @classmethod
    def get_request_fields(cls):
        return list(filter(lambda property: isinstance(property, RequestField),
                           cls.get_request_properties()))

    @classmethod
    def get_resource_fields(cls):
        return list(filter(lambda property: isinstance(property, ResourceField),
                           cls.get_request_properties()))

    @classmethod
    def get_required_request_fields(cls):
        return list(filter(lambda property: isinstance(property, RequestField),
                           cls.get_required_request_properties()))

    @classmethod
    def get_tasks(cls):
        endpoint_tasks = filter(lambda property: isinstance(property, EndpointTask),
                                cls.get_endpoint_attributes())

        return sorted(endpoint_tasks, key=lambda task: task.priority)

    @classmethod
    def get_url_fields(cls):
        return list(filter(lambda property: isinstance(property, RequestUrlField),
                           cls.get_endpoint_attributes()))

    @classmethod
    def documentation(cls):
        return {
            'class_name': cls.__name__,
            'fields': [p.documentation for p in cls.get_request_properties()],
        }

    @classmethod
    def get_adhoc_queries(cls):
        return [
            prop for prop in cls.get_endpoint_attributes() if isinstance(prop, RequestAdhocQuerySet)
        ]


class EndpointDefinition(BaseEndpointDefinition):
    """ A base class to be used when defining endpoints.

    Base class to be used implementing endpoints that aren't necessarily tied to a model. Also implements
    basic consumer-based authentication.
    """
    request = RawRequestObjectProperty()
    _consumer_type = ConsumerAttribute(field_name='type', default='RW')
    is_read_only = False
    """ Used to determine accessibility for the current consumer.
    """

    def is_permitted(self):
        """ Checks authorization for the current consumer.

        Returns:
            ``bool``: Whether or not the user has permission to the resource.
        """
        if self._consumer_type == None or self._consumer_type == BaseConsumer.TYPE_READ_WRITE:
            return True
        if self._consumer_type == BaseConsumer.TYPE_READ_ONLY:
            if self.is_read_only:
                return True
            if self.request.method == 'GET':
                return True
            else:
                self._validation_error_message = 'Action not allowed for read-only consumer'
                return False
        return False

    @classmethod
    def get_consumer_attributes(cls):
        return list(filter(lambda property: isinstance(property, ConsumerAttribute),
                           cls.get_request_properties()))

    @classmethod
    def get_consumer_type(cls):
        consumer_attribute = cls.get_consumer_attributes()
        if len(consumer_attribute) == 1:
            consumer_attribute = consumer_attribute[0]
            return consumer_attribute.name
        else:
            return 'unknown'

    @classmethod
    def documentation(cls):
        docs = super().documentation()
        docs['consumer_type'] = cls.get_consumer_type()
        return docs


class ResourceCreationMixin(object):
    @property
    def http_status(self):
        return http.client.CREATED


class ResourceEndpointDefinition(EndpointDefinition):
    """ A base class to be used when defining endpoints bound to models.
    """
    consumer = RequestAttribute()

    resource_id = RequestUrlField(name='id', description="UUID of the resource to retrieve")
    """ The ID of the resource being fetched or updated.
    """
    resource_model = None
    """ The model to attach to the resource endpoint definition.

    Must extend or implement the Django ORM model interface as required.
    """

    def __init__(self, *args, **kwargs):
        super(ResourceEndpointDefinition, self).__init__()
        self._cached_resource = None

    @property
    def resource(self):
        """ Resource implementation

        Queries the object manager of `self.resource_model` for the given id (`self.resource_id`).
        """
        if not self._cached_resource:
            self._cached_resource = self.resource_model.objects.get(id=self.resource_id)
        return self._cached_resource


class ResourceUpdateEndpointDefinition(ResourceEndpointDefinition):
    @EndpointTask(priority=-100)
    def mutate(self):
        resource = self.resource
        for resource_field in self.get_resource_fields():
            field_value = getattr(self, resource_field.name)
            if field_value is not None:
                setattr(resource, resource_field.name, field_value)

    @EndpointTask(priority=-101)
    def validate_input(self):
        resource = self.resource
        expected_fields = set(list(field.name for field in self.get_resource_fields()) + list(
            field.name for field in self.get_request_fields()))
        unexpected = self.request.body_field_names - expected_fields
        if unexpected:
            raise errors.ClientErrorUnprocessableEntity('Unexpected fields: {}'.format(', '.join(unexpected)))


task = EndpointTask
deferrable_task = DeferrableEndpointTask
request_attribute = RequestAttribute
consumer_attribute = ConsumerAttribute
field = RequestField
resource_field = ResourceField
url_field = RequestUrlField
adhoc_queryset = RequestAdhocQuerySet
aggregate = Aggregate
require_one = RequireOneAttribute
require_all = RequireAllAttribute
require_all_if_any = RequireAllIfAnyAttribute
endpoint_resource = EndpointResourceAttribute
endpoint_response = EndpointResponseAttribute



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

from collections import defaultdict
import inspect
import types

from django.db import models
from django.db.models import ManyToOneRel

NEVER = 0
ALWAYS = 1
IF_TRUTHY = 2

DEFAULT_UNEXPANDED_VALUE = object()
EXPANDABLE_FIELD_KEY = '__expandable__'


class _ExpandableForeignKey():
    def __init__(self, display_key, model_class):
        self.display_key = display_key
        self.model_class = model_class


def expandable(model_class=None, display_key=None):
    if model_class and not issubclass(model_class, (models.Model,)):
        raise ValueError('model_class must be an instance of a Django Model')
    if model_class and display_key:
        try:
            model_class._meta.get_field(display_key)
        except models.FieldDoesNotExist as e:
            raise ValueError(f'{display_key} is not a field on {model_class.__name__}')
    return _ExpandableForeignKey(display_key, model_class)


def _get_unexpanded_field_value(inst, field_name, field_type):
    if not field_type.model_class:
        return DEFAULT_UNEXPANDED_VALUE

    display_key = field_type.display_key or field_type.model_class._meta.pk.name
    is_multiple = isinstance(inst.__class__._meta.get_field(field_name), ManyToOneRel)

    if is_multiple:
        # special case for keys that have multiple values (for instance, inverse fk relations)
        obj = getattr(inst, field_name)
        return [{display_key: getattr(v, display_key)} for v in obj.all()]

    if display_key == field_type.model_class._meta.pk.name:
        # special case - we know this is a primary key, so we can get it without retrieving the object
        return {display_key: getattr(inst, field_name + '_id')}
    else:
        # we're not returning the PK - have to actually retrieve the model
        obj = getattr(inst, field_name)
        return {display_key: getattr(obj, display_key)}


def _get_filtered_field_value(inst, field_name, field_type, filter_def, expand_this, expand_children):
    # get the value from inst
    if field_type == NEVER:
        return None

    if isinstance(field_type, types.FunctionType):
        val = field_type(inst)
    elif isinstance(field_type, _ExpandableForeignKey) and not expand_this:
        val = _get_unexpanded_field_value(inst, field_name, field_type)
    else:
        try:
            val = getattr(inst, field_name)
        except (AttributeError, models.fields.FieldDoesNotExist) as e:
            return None

    if isinstance(val, models.Manager):
        # need to get an iterable to proceed
        val = val.all()

    # should this value be passed through the filters itself?
    if val.__class__ in filter_def or isinstance(val, (list, tuple, models.Model, models.query.QuerySet)):
        val = _apply_filters_to_object(val, filter_def,
                                       expand_children=expand_children,
                                       klass=val.__class__)

    if (field_type == ALWAYS) or \
            isinstance(field_type, types.FunctionType) or\
            ((field_type == IF_TRUTHY) and val) or \
            isinstance(field_type, _ExpandableForeignKey):
        return val
    else:
        return None


def _apply_filters_to_object(inst, filter_def, expand_children=None, klass=None):
    if isinstance(inst, (list, tuple, models.query.QuerySet)):
        # if it's a tuple or list, iterate over the collection and call _apply_filters_to_object on each item
        return [_apply_filters_to_object(item, filter_def, expand_children=expand_children, klass=item.__class__)
                for item in inst]

    fields_def = filter_def.get(klass)
    if fields_def is None:
        # no filter definition for this class.  Check the inheritance chain to see if we have something there
        result = None
        for base_class in inspect.getmro(klass):
            if base_class in filter_def:
                result = _apply_filters_to_object(inst, filter_def,
                                                  expand_children=expand_children,
                                                  klass=base_class)
                break
        return result
    else:
        # first, recursively populate from any ancestor classes in the inheritance hierarchy
        mro = inspect.getmro(klass)
        result = defaultdict(list)
        if len(mro) > 1:
            filtered_ancestor = _apply_filters_to_object(inst, filter_def,
                                                         expand_children=expand_children,
                                                         klass=mro[1])
            if filtered_ancestor:
                result.update(filtered_ancestor)

        # now populate the fields according to fields defined in filter_def
        expandables = []
        if isinstance(fields_def, types.FunctionType):
            result = fields_def(inst)
        elif fields_def is ALWAYS:
            return inst
        else:
            for field_name, field_type in fields_def.items():
                if isinstance(field_type, _ExpandableForeignKey):
                    expandables.append(field_name)

                value = _get_filtered_field_value(inst, field_name, field_type, filter_def,
                                                  expand_this=field_name in expand_children,
                                                  expand_children=expand_children.get(field_name, {}))

                if value is not None and value != DEFAULT_UNEXPANDED_VALUE:
                    result[field_name] = value
                elif (field_type == NEVER) and (field_name in result):
                    del result[field_name]

        if expandables:
            result[EXPANDABLE_FIELD_KEY] += expandables

        return result


def _compile_expansion(expand_fields):
    top = {}
    for field in expand_fields:
        d = top
        for key in field.strip().split('.'):
            d.setdefault(key, {})
            d = d[key]
    return top


def apply_filters_to_object(inst, filter_def, expand_header=''):
    if expand_header:
        expand_dict = _compile_expansion(expand_header.split(','))
    else:
        expand_dict = {}
    return _apply_filters_to_object(inst, filter_def, expand_children=expand_dict, klass=inst.__class__)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import json
import time
import urllib.parse
import logging

import oauth2

from oauthlib import oauth1 as oauthlib_oauth1
from oauthlib import common as oauthlib_common

from django.http import QueryDict

import django.test
from django.test import Client
from django.test.client import ClientHandler
from django.test.runner import DiscoverRunner

from django_declarative_apis import models

DEFAULT_CONTENT_TYPE = 'application/x-www-form-urlencoded'
_ENCODERS = {
    DEFAULT_CONTENT_TYPE: lambda data: urllib.parse.urlencode(data),
    'application/json': lambda data: json.dumps(data),
}


class NoLoggingTestRunner(DiscoverRunner):

    """Don't log during tests."""

    def run_tests(self, test_labels, extra_tests=None, **kwargs):
        logging.disable(logging.CRITICAL)
        return super(NoLoggingTestRunner, self).run_tests(test_labels, extra_tests, **kwargs)


class RequestCreatorMixin:
    def setUp(self):
        self.consumer = models.OauthConsumer.objects.create(name='test')
        self.consumer.save()

    def create_request(self, method='GET', path='/', url_fields=None, body=None, content_type=DEFAULT_CONTENT_TYPE,
                       use_auth_header_signature=False):
        meth = getattr(django.test.RequestFactory(), method.lower())
        req = meth(
            path,
            _ENCODERS[content_type](body) if body else None,
            content_type=content_type
        )
        req.consumer = self.consumer
        req.META['consumer'] = self.consumer

        if url_fields is not None:
            params = req.GET.copy()
            params.update(url_fields)
            req.GET = params

        # TODO: this should be better documented or on by default. If this is omitted or False, then requests will
        # fail in resource input validation (ResourceUpdateEndpointDefinition.validate_input)
        req.META['use_auth_header_signature'] = use_auth_header_signature
        OAuthClientHandler._build_request(req)

        return req


class OAuthClientHandler(ClientHandler):
    one_time_client_timestamp_override = None

    @classmethod
    def get_oauth_client_timestamp(cls):
        if cls.one_time_client_timestamp_override:
            result = cls.one_time_client_timestamp_override
            cls.one_time_client_timestamp_override = None
        else:
            result = int(time.time())
        return result

    @classmethod
    def _build_request(cls, request):
        consumer = request.META.get('consumer')

        # Make request params mutable so we can add authorization parameters.
        # We make the params immutable again before processing the request
        request.POST = request.POST.copy()
        request.GET = request.GET.copy()

        if consumer:
            if request.method == 'POST':
                data = request.POST
            else:
                data = request.GET

            # This provides a way for us to override default values for testing.
            oauth_version = request.META.get('oauth_version', '1.0')
            oauth_nonce = request.META.get('oauth_nonce', oauth2.generate_nonce())
            oauth_client_timestamp = request.META.get('oauth_timestamp', cls.get_oauth_client_timestamp())

            rsa_key = request.META.get('rsa_key', None)
            oauth_signature_method = oauthlib_oauth1.SIGNATURE_RSA if rsa_key else oauthlib_oauth1.SIGNATURE_HMAC

            oauth_signature_data = {'oauth_version': oauth_version,
                                    'oauth_nonce': oauth_nonce,
                                    'oauth_timestamp': str(oauth_client_timestamp),
                                    'oauth_consumer_key': consumer.key,
                                    'oauth_signature_method': oauth_signature_method}

            # collect ALL request parameters (original + OAuth) for signing
            all_request_parameters = data.copy()
            all_request_parameters.update(oauth_signature_data)

            if oauth_signature_method == oauthlib_oauth1.SIGNATURE_RSA:
                # use RSA-SHA1 signature method
                oauth1_client = oauthlib_oauth1.Client(
                    consumer.key,
                    signature_method=oauth_signature_method,
                    rsa_key=rsa_key)

                oauth_request = oauthlib_common.Request(request.build_absolute_uri(request.path),
                                                        http_method=request.method,
                                                        body=all_request_parameters)

                oauth_signature_data['oauth_signature'] = oauth1_client.get_oauth_signature(oauth_request)
            else:
                # use HMAC-SHA1 signature method
                oauth_signature_data.update({'oauth_signature_method': 'HMAC-SHA1'})

                # Create oauth request object to compute signature
                oauth_request = oauth2.Request.from_consumer_and_token(consumer, None,
                                                                       request.method,
                                                                       request.build_absolute_uri(request.path),
                                                                       all_request_parameters,
                                                                       is_form_encoded=True)

                # Add signature to django request
                signature_method = oauth2.SignatureMethod_HMAC_SHA1()
                oauth_request.sign_request(signature_method, consumer, None)
                oauth_signature_data['oauth_signature'] = oauth_request.get_parameter('oauth_signature')

            use_auth_header_signature = request.META.pop('use_auth_header_signature', False)
            if use_auth_header_signature:
                auth_header_string = "OAuth " + ','.join(
                    ['{0}="{1}"'.format(key, value) for key, value in oauth_signature_data.items()])
                request.META['HTTP_AUTHORIZATION'] = auth_header_string
            else:
                data.update(oauth_signature_data)

        # Recreate the GET and POST QueryDicts to make them immutable, as in production
        request.POST = QueryDict(request.POST.urlencode().encode('utf-8'))
        request.GET = QueryDict(request.GET.urlencode().encode('utf-8'))

    def get_response(self, request):
        self._build_request(request)
        return ClientHandler.get_response(self, request)


class DeclarativeApisOAuthClient(Client):
    def __init__(self, *args, **kwargs):
        super(DeclarativeApisOAuthClient, self).__init__(*args, **kwargs)
        self.handler = OAuthClientHandler()

    def request(self, expected_status_code=None, **kwargs):
        response = super(DeclarativeApisOAuthClient, self).request(**kwargs)
        if expected_status_code is not None:
            if response.status_code != expected_status_code:
                print(response.content.decode('utf-8'))
                raise AssertionError(
                    'Status code, {}, did not match expected status code, {}'.format(response.status_code,
                                                                                     expected_status_code))
        return response

    def put(self, path, data=None, expected_status_code=None, **kwargs):
        if 'content_type' not in kwargs:
            data = data or {}
            kwargs['content_type'] = 'application/x-www-form-urlencoded'
            data_qd = QueryDict(mutable=True)
            data_qd.update(data)
            data = data_qd.urlencode()

        return super(DeclarativeApisOAuthClient, self).put(path, data, expected_status_code=expected_status_code, **kwargs)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import http
import unittest
import mock
import django.test
from django_declarative_apis import adapters, machinery, authentication


class BaseHandlerTestCase(unittest.TestCase):
    def test_supported_methods(self):
        self.assertEqual(
            adapters.BaseHandler.defined_methods,
            {'get', 'put', 'patch', 'post', 'delete'}
        )


class _HandlerA(machinery.EndpointDefinition):
    RESPONSE = {'foo': 'bar'}
    this_is_really_a_field = machinery.field(type=str)

    def is_authorized(self):
        return True

    @property
    def resource(self):
        return self.RESPONSE


class _HandlerB(machinery.EndpointDefinition):
    def is_authorized(self):
        return True

    @property
    def resource(self):
        return {}


class EndpointHandlerTestCase(unittest.TestCase):
    def test_invalid_method_error(self):
        self.assertRaises(TypeError, adapters.EndpointHandler, invalid=lambda: None)

    def test_multiple_handlers(self):
        endpoints = (_HandlerA, _HandlerB)
        handler = adapters.EndpointHandler(get=endpoints)
        self.assertEqual(handler.method_handlers['GET'].endpoint_definitions, endpoints)
        self.assertEqual(handler.allowed_methods, {'GET',})

    def test_single_handler(self):
        handler = adapters.EndpointHandler(get=_HandlerA)
        self.assertEqual(handler.method_handlers['GET'].endpoint_definitions, (_HandlerA,))
        self.assertEqual(handler.allowed_methods, {'GET',})

    def test_documentation(self):
        handler = adapters.EndpointHandler(get=_HandlerA)
        docs = handler.documentation
        self.assertIn('GET', docs)
        self.assertIn('class_name', docs['GET'][0])
        self.assertEqual(docs['GET'][0]['class_name'], '_HandlerA')
        self.assertIn('fields', docs['GET'][0])

        expected_field = {
            'name': 'this_is_really_a_field',
            'type': str,
            'multivalued': False
        }
        self.assertIn(expected_field, docs['GET'][0]['fields'])

    def test_handle_request(self):
        req = django.test.RequestFactory().get('/simple')
        req.consumer = None
        resp = adapters.EndpointHandler(get=_HandlerA).handle_request('GET', req)
        self.assertEqual(resp, (http.HTTPStatus.OK, _HandlerA.RESPONSE))

    def test_custom_request_field(self):
        class _HandlerWithoutCustomField(machinery.EndpointDefinition):
            def is_authorized(self):
                try:
                    return self.request.custom_request_field == 42
                except AttributeError:
                    return False

            @property
            def resource(self):
                return {}

        class _HandlerWithCustomField(_HandlerWithoutCustomField):
            request = machinery.RawRequestObjectProperty(additional_safe_fields=('custom_request_field',))

        class _Authenticator(authentication.Authenticator):
            def is_authenticated(self, request):
                request.custom_request_field = 42
                return True

            def challenge(self, error):
                pass

        authenticator = _Authenticator()
        req = django.test.RequestFactory().get('/simple')
        req.consumer = None
        resource = adapters.EndpointResource(get=_HandlerWithoutCustomField, authentication={None: [authenticator,]})
        resp = resource(req)
        self.assertEqual(resp.status_code, http.HTTPStatus.FORBIDDEN)

        resource = adapters.EndpointResource(get=_HandlerWithCustomField, authentication={None: [authenticator,]})
        resp = resource(req)
        self.assertEqual(resp.status_code, http.HTTPStatus.OK)


class EndpointResourceTestCase(unittest.TestCase):
    def test_custom_authentication(self):
        class _Authenticator(authentication.Authenticator):
            def is_authenticated(self, request):
                return True

            def challenge(self):
                pass

        authenticator = _Authenticator()
        resource = adapters.EndpointResource(authentication={None: [authenticator,]})
        self.assertEqual(resource.authentication, {None: [authenticator]})



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'db.sqlite3'
    }
}

SECRET_KEY = 'NOT A REAL KEY'

DEBUG = True

INSTALLED_APPS = [
    'django_declarative_apis',
    'django.contrib.contenttypes',
    'tests',
]

ROOT_URLCONF = 'tests.urls'

TEST_RUNNER = 'tests.testutils.NoLoggingTestRunner'

MIDDLEWARE = []

REQUIRE_HTTPS_FOR_OAUTH = False

DECLARATIVE_ENDPOINT_RESOURCE_ADAPTER = 'django_declarative_apis.adapters.EndpointResource'
DECLARATIVE_ENDPOINT_DEFAULT_FILTERS = 'tests.filters.DEFAULT_FILTERS'
DECLARATIVE_ENDPOINT_AUTHENTICATION_HANDLERS = (
    (
        (None, 'django_declarative_apis.authentication.oauthlib.oauth1.TwoLeggedOauth1Hint',),
        'django_declarative_apis.authentication.oauthlib.oauth1.TwoLeggedOauth1',
    ),
)
DECLARATIVE_ENDPOINT_TASKS_FORCE_SYNCHRONOUS = False
DECLARATIVE_ENDPOINT_TASKS_SYNCHRONOUS_FALLBACK = False
